import typing
from typing import TYPE_CHECKING, Optional, Union

from onetick.py.otq import otq
from onetick.py.core.column_operations.base import _Operation
from onetick.py.backports import Literal
from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


DATA_QUALITY_EVENTS = Literal[
    'COLLECTOR_FAILURE',
    'EMPTY',
    'MISSING',
    'MOUNT_BAD',
    'OK',
    'PATCHED',
    'QUALITY_DELAY_STITCHING_WITH_RT',
    'QUALITY_OK_STITCHING_WITH_RT',
    'STALE',
]


@inplace_operation
def show_data_quality(self: 'Source', inplace=False) -> Optional['Source']:
    """
    This method shows data quality events within the interval of the query.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    | **SHOW_DATA_QUALITY** OneTick event processor
    | :py:meth:`insert_data_quality_event`
    | :py:meth:`intercept_data_quality`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    By default data quality events are not showed, use this method to see them:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.insert_data_quality_event('OK')
    >>> data = data.show_data_quality()
    >>> otp.run(data)
                         Time  DATA_QUALITY_TYPE DATA_QUALITY_NAME
    0 2003-12-01 00:00:00.000                  0                OK
    1 2003-12-01 00:00:00.001                  0                OK
    2 2003-12-01 00:00:00.002                  0                OK
    """

    self.sink(
        otq.ShowDataQuality()
    )
    self.schema.set(**{'DATA_QUALITY_TYPE': int, 'DATA_QUALITY_NAME': str})
    return self


@inplace_operation
def insert_data_quality_event(
    self: 'Source',
    data_quality: DATA_QUALITY_EVENTS,
    where: Optional[Union[str, _Operation]] = None,
    insert_before: bool = True,
    inplace=False,
) -> Optional['Source']:
    """
    Insert data quality events into the data flow.

    Parameters
    ----------
    data_quality: str
        The type of data quality event.
        Must be one of the supported data quality event types.
    where: str or :class:`Operation`
        Specifies a criterion for the selection of ticks whose arrival results in generation of a data quality event.
        If this expression returns True, a data quality event will be inserted into a time series.
        The expression can also be empty (default), in which case each input tick generates a data quality event.
    insert_before: bool
        If True (default),
        generated data quality event tick will be inserted before the input tick that triggered its creation,
        otherwise generated data quality event will be inserted after that input tick.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    | **INSERT_DATA_QUALITY_EVENT** OneTick event processor
    | :py:meth:`show_data_quality`
    | :py:meth:`intercept_data_quality`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------
    Insert OK data quality event before each tick:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.insert_data_quality_event('OK')
    >>> otp.run(data)
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    By default data quality events are not showed,
    use :py:meth:`show_data_quality` method to see them:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.insert_data_quality_event('OK')
    >>> data = data.show_data_quality()
    >>> otp.run(data)
                         Time  DATA_QUALITY_TYPE DATA_QUALITY_NAME
    0 2003-12-01 00:00:00.000                  0                OK
    1 2003-12-01 00:00:00.001                  0                OK
    2 2003-12-01 00:00:00.002                  0                OK

    Use ``where`` parameter to specify the condition when to insert ticks:

    >>> data = otp.Ticks({'SIZE': [1, 200, 3]})
    >>> data = data.insert_data_quality_event('MISSING', where=data['SIZE'] > 100)
    >>> data = data.show_data_quality()
    >>> otp.run(data)
                         Time  DATA_QUALITY_TYPE DATA_QUALITY_NAME
    0 2003-12-01 00:00:00.001                  2           MISSING
    """
    if data_quality not in typing.get_args(DATA_QUALITY_EVENTS):
        raise ValueError(f"Parameter 'data_quality' doesn't support value: {data_quality}")

    self.sink(
        otq.InsertDataQualityEvent(
            data_quality=data_quality,
            where=str(where) if where is not None else '',
            insert_before=insert_before,
        )
    )

    return self


@inplace_operation
def intercept_data_quality(self: 'Source', inplace=False) -> Optional['Source']:
    """
    This method removes data quality messages, thus preventing them from delivery to the client application.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    | **INTERCEPT_DATA_QUALITY** OneTick event processor
    | :py:meth:`insert_data_quality_event`
    | :py:meth:`show_data_quality`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    By default data quality events are not showed, use this method to see them:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.insert_data_quality_event('OK')
    >>> data = data.show_data_quality()
    >>> otp.run(data)
                         Time  DATA_QUALITY_TYPE DATA_QUALITY_NAME
    0 2003-12-01 00:00:00.000                  0                OK
    1 2003-12-01 00:00:00.001                  0                OK
    2 2003-12-01 00:00:00.002                  0                OK

    Intercepting data quality events will remove them from the data flow:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.insert_data_quality_event('OK')
    >>> data = data.intercept_data_quality()
    >>> data = data.show_data_quality()
    >>> otp.run(data)  # doctest: +ELLIPSIS
    Empty DataFrame
    ...
    """
    self.sink(
        otq.InterceptDataQuality()
    )
    return self


@inplace_operation
def show_symbol_errors(self: 'Source', inplace=False) -> Optional['Source']:
    """
    This method propagates a tick representing per-symbol error.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    | **SHOW_SYMBOL_ERRORS** OneTick event processor
    | :py:meth:`intercept_symbol_errors`
    | :py:meth:`onetick.py.Source.throw`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    By default symbol errors are not showed, use this method to see them:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.throw('WRONG', scope='symbol')
    >>> data = data.show_symbol_errors()
    >>> otp.run(data)
            Time  ERROR_CODE ERROR_MESSAGE
    0 2003-12-01           1         WRONG
    1 2003-12-01           1         WRONG
    2 2003-12-01           1         WRONG
    """

    self.sink(
        otq.ShowSymbolErrors()
    )
    self.schema.set(**{'ERROR_CODE': int, 'ERROR_MESSAGE': str})
    return self


@inplace_operation
def intercept_symbol_errors(self: 'Source', inplace=False) -> Optional['Source']:
    """
    This method removes removes per-symbol errors, thus preventing them from delivery to the client application.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    | **INTERCEPT_SYMBOL_ERRORS** OneTick event processor
    | :py:meth:`show_symbol_errors`
    | :py:meth:`onetick.py.Source.throw`

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    By default data quality events are not showed, use this method to see them:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.throw('WRONG', scope='symbol')
    >>> data = data.show_symbol_errors()
    >>> otp.run(data)
            Time  ERROR_CODE ERROR_MESSAGE
    0 2003-12-01           1         WRONG
    1 2003-12-01           1         WRONG
    2 2003-12-01           1         WRONG

    Intercepting data quality events will remove them from the data flow:

    >>> data = otp.Ticks({'A': [1, 2, 3]})
    >>> data = data.throw('WRONG', scope='symbol')
    >>> data = data.intercept_symbol_errors()
    >>> data = data.show_symbol_errors()
    >>> otp.run(data)  # doctest: +ELLIPSIS
    Empty DataFrame
    ...
    """
    self.sink(
        otq.InterceptSymbolErrors()
    )
    return self



import os
from inspect import getframeinfo, stack
from onetick.py.otq import otq

_graph_components = otq.graph_components
_internal_utils = otq._internal_utils


def _get_reference_counted_prefix():
    try:
        return _internal_utils.get_reference_counted_prefix()
    except AttributeError:
        return _internal_utils.get_referance_counted_prefix()


class BaseOqdEp(_graph_components.EpBase):  # type: ignore[name-defined]
    """
    Base class for all OQD EPs
    """
    class Parameters:
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["shared_state_variables"]
            return list_val

    __slots__ = ["shared_state_variables", "_default_shared_state_variables", "stack_info", "_used_strings"]

    def __init__(self, shared_state_variables=""):
        _graph_components.EpBase.__init__(self, self._get_name())
        self._default_shared_state_variables = ""
        self.shared_state_variables = shared_state_variables
        self._used_strings = {}
        for param_name in type(self).__dict__:
            param_val = getattr(self, param_name, '')
            if (
                isinstance(param_val, str) and
                _get_reference_counted_prefix() in param_val and
                param_val not in self._used_strings
            ):
                _internal_utils.inc_ref_count(param_val)
                self._used_strings[param_val] = 1
        caller = getframeinfo(stack()[1][0])
        self.stack_info = caller.filename + ":" + str(caller.lineno)

    def _get_name(self):
        raise NotImplementedError()

    def set_shared_state_variables(self, value):
        self.shared_state_variables = value
        return self

    def __repr__(self):
        return self._to_string(for_repr=True)

    def __str__(self):
        return self._to_string()

    def __del__(self):
        for param_name in self._used_strings:
            _internal_utils.dec_ref_count(param_name)
            if _internal_utils.get_ref_count(param_name) == 0:
                _internal_utils.remove_from_memory(param_name)

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc


class OqdSourceDprcMain(BaseOqdEp):
    """
    OQD_SOURCE_DPRC_MAIN

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve a time series of unadjusted
    prices for a symbol for the main pricing exchange.

    The OQD_SOURCE_DPRC_MAIN EP retrieves a single daily time series
    corresponding to the main pricing line for a symbol.
    OQD_SOURCE_DPRC_MAIN differs from OQD_SOURCE_DPRC_EXCH in that it
    doesn't require specification of an exchange.  The most appropriate
    exchange will be chosen by the EP.

    The main pricing line is defined as the composite price for countries
    where a composite price is available.  If no composite price is
    available, then the exchange of primary listing is used, except for
    Germany where XETRA is chosen as the main pricing source.

    To adjust the data for corporate actions, use either the
    OQD_CORP_ACTION EP or the CORP_ACTION EP.

    To retrieve a price series for only one particular exchange, see
    OQD_SOURCE_DPRC_EXCH.  For all available price series, see
    OQD_SOURCE_DPRC_ALL.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_DPRC_MAIN.otq"


class OqdCorpAction(BaseOqdEp):
    """
    OQD_CORP_ACTION

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to generate a sparse multiplicative
    adjustment series for use in queries requiring explicit corporate
    action adjustments.

    This EP is used for writing queries that require finer control over
    the corporate action adjustment process than is provided by the
    CORP_ACTION EP or for queries that don't set the symbol date and
    utilize the Onetick Reference File.

    For a simpler method of applying corporate action adjustments, see the
    CORP_ACTION EP.

    When adjusting historical data for corporate
    actions using multiplicative factors, each data point is adjusted by
    all future corporate actions. This means that multiplicative corporate
    action adjustments are cumulative going backwards in time.

    The OQD_CORP_ACTION EP generates the correct time series of adjustment factors
    for use in a query utilizing the JOIN_BY_TIME EP.  The output of this
    EP is guaranteed to start at the query start time.  If a symbol has no
    corporate actions over the queried time range, this EP will return a
    single tick at the query start time with a value of 1.0 for the
    adjustment factors.

    Price and volume adjustments are applied inversely.  Each tick from
    the OQD_CORP_ACTION EP will return 2 fields  of adjustment factors: MADJ_PRICE and
    MADJ_VOLUME.

    To adjust prices or other data items stated on a per share
    basis, multiple the data series by the value in the MADJ_PRICE column.

    To adjust volume or other data items stated on a share basis, multiple
    the data series by the MADJ_VOLUME column.  The factors for spinoff are
    not include in the MADJ_VOLUME factor.

    The typical usage of this EP is to pass the output of the EP into a
    JOIN_BY_TIME node to join the adjustment factors with a data series
    that needs adjustments applied, and then perform the adjustments in
    and ADD_FIELD or UPDATE_FIELD EP.

    Because the OQD_CORP_ACTION EP goes directly to the OneQuantData
    databases, it does not require that a Onetick reference file has been
    configured and loaded and may be used even when SYMBOL_DATE is set to 0.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters:


    ADJUST_TYPES
            (string)



    A string containing a comma separate list of adjustment types to
    apply.  The types may be any of: SPLIT, STOCK_DIVIDEND, RIGHTS, or
    SPINOFF.  The default is SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF.
    """

    class Parameters:
        adjust_types = "ADJUST_TYPES"
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["adjust_types", "shared_state_variables"]
            return list_val

    __slots__ = ["adjust_types", "_default_adjust_types"]

    def __init__(self, adjust_types="SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF", shared_state_variables=""):
        BaseOqdEp.__init__(self, shared_state_variables)
        self._default_adjust_types = "SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF"
        self.adjust_types = adjust_types

    def set_adjust_types(self, value):
        self.adjust_types = value
        return self

    def _get_name(self):
        return "OTQ::OQD_CORP_ACTION.otq"

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.adjust_types != "SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF":
            desc += "adjust_types=" + py_to_str(self.adjust_types) + ","
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc


class OqdCorpDailyFactor(BaseOqdEp):
    """
    OQD_CORP_DAILY_FACTOR

    &#160;

    Type: Other

    Description:A
    OneQuantData&#153; source EP to generate a daily series of
    multiplicative corporate action adjustment factors.

    This EP will generate a time series containing the total
    multiplicative adjustment factors for each day over the query range.
    For days on which there were no corporate actions, the
    OQD_CORP_DAILY_FACTOR EP will return values of 1.0.

    For adjusting the entire history of a time series, see the CORP_ACTION
    EP or the OQD_CORP_ACTION EP.

    The OQD_CORP_DAILY_FACTOR EP is used when a time series is needed that
    calculates the daily adjustment factor taking into account the possibility
    of multiple corporate actions on the same day.  This EP only
    accumulates factors per day.  It does not accumulate factors over
    time.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters:


    ADJUST_TYPES (string)


    A string containing a comman separate list of adjustment types to
    apply.  The types may be any of: SPLIT, STOCK_DIVIDEND, RIGHTS, or
    SPINOFF.  The default is SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF.
    """

    class Parameters:
        action_type = "ACTION_TYPE"
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["action_type", "shared_state_variables"]
            return list_val

    __slots__ = ["action_type", "_default_action_type"]

    def __init__(self, action_type="SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF", shared_state_variables=""):
        BaseOqdEp.__init__(self, shared_state_variables)
        self._default_action_type = "SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF"
        self.action_type = action_type

    def set_action_type(self, value):
        self.action_type = value
        return self

    def _get_name(self):
        return "OTQ::OQD_CORP_DAILY_FACTOR.otq"

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.action_type != "SPLIT,STOCK_DIVIDEND,RIGHTS,SPINOFF":
            desc += "action_type=" + py_to_str(self.action_type) + ","
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc


class OqdSourceBbgbsym(BaseOqdEp):
    """
    OQD_SOURCE_BBGBSYM

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology
    using a symbol in the full Bloomberg ticker format, {TICKER} {EXCH_CODE} {Yellow Key} .

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
    Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
    share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGBSYM EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
    on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGBSYM through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGBSYM.otq"


class OqdSourceBbgbtkr(BaseOqdEp):
    """
    OQD_SOURCE_BBGBTKR

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology using a symbol
    in the short Bloomberg ticker format, {TICKER} {EXCH_CODE} .

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
     Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
     share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGBTKR EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
    on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGBTKR through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGBTKR.otq"


class OqdSourceBbgfgc(BaseOqdEp):
    """
    OQD_SOURCE_BBGFGC

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology
    using a symbol that is a Composite FIGI.

    Multiple FIGI identifiers may be associated with a single composite FIGI.
    This EP will return multiple ticks per day for each FIGI associated with the composite FIGI.

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
    Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
    share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGFGC EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
    on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGFGC through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGFGC.otq"


class OqdSourceBbgfgs(BaseOqdEp):
    """
    OQD_SOURCE_BBGFGS

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology
    using a symbol that is a share class FIGI.

    Multiple FIGI identifiers may be associated with a single share class FIGI.
    This EP will return multiple ticks per day for each FIGI associated with the share class FIGI.
    The share class FIGI may be used to find alternate venues where a security trades.

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
    Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
    share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGFGS EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
    on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGFGS through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGFGS.otq"


class OqdSourceBbgfgv(BaseOqdEp):
    """
    OQD_SOURCE_BBGFGV

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology
    using a symbol that is a venue FIGI, which is the FIGI associated
    with a particular trading line on a particular venue.

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
    Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
    share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGFGV EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
     on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGFGV through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGFGV.otq"


class OqdSourceBbgoid(BaseOqdEp):
    """
    OQD_SOURCE_BBGOID

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve OpenFIGI symbology
    using a symbol that is a OneMarketData OID.

    Multiple FIGI identifiers may map to a single OID.
    This EP may return multiple ticks per day for each FIGI associated with an OID.

    This EP will return a time series that includes the OneMarketData OID, the venue FIGI,
    Bloomberg yellow key, Bloomberg exchange code, Bloomberg ticker, composite FIGI,
    share class FIGI, Bloomberg security type, and Bloomberg security name.

    Coverage for OpenFIGI symbology begins on October 6th, 2014.

    The OQD_SOURCE_BBGOID EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day. After the first tick, there will only be ticks
     on days when some field in the data changes.

    Pass the output of OQD_SOURCE_BBGOID through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    The tick type for this EP may be set to OQD::*
    The correct database and tick type will be supplied by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_BBGOID.otq"


class OqdSourceCacs(BaseOqdEp):
    """
    OQD_SOURCE_CACS

    &#160;

    Type: Other

    Description:A
    OneQuantData&#153; source EP to retrieve a time series of corporate
    actions for a symbol.

    This EP will return all corporate action fields available for a symbol
    with EX-Dates between the query start time and end time.  The
    timestamp of the series is equal to the EX-Date of the corporate
    action with a time of 0:00:00 GMT.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_CACS.otq"


class OqdSourceCact(BaseOqdEp):
    """
    OQD_SOURCE_CACT
     &#160;
     Type: Other
     Description: A
    OneQuantData&#153; source EP to retrieve a time series of corporate
    actions for all symbols.

    This EP will return a subset of corporate action fields for all
    symbols.

    The OQD_SOURCE_CACT EP is used in conjunction with the special symbols EX_DATE,
    ANN_DATE, PAY_DATE, and REC_DATE.  It will return all corporate
    actions for all symbols with the specified type of date between the query
    start time and query end time.

    For example, if the special symbol EX_DATE is used, then the timestamp
    of the series is the corporate action EX-date with a time of 0:00:00
    GMT and the EP will return all corporate actions with EX-dates between
    the query start time and the query end time.  If the symbol is
    REC_DATE, then all of the timestamps would correspond to corporate
    action record dates.

    The OQD_SOURCE_CACT EP is typically used in stage 1 queries to quickly
    discover a list of securities involved in corporate actions.  Please
    note that if using one of the special symbols other than EX_DATE, the
    EX-date of the corporate action may fall outside of the query time
    range.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_CACT.otq"


class OqdSourceDes(BaseOqdEp):
    """
    OQD_SOURCE_DES

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve a time series of
    descriptive fields for a symbol.

    This EP will return the history of descriptive fields for a symbol.
    There will only be ticks on days when some field in the descriptive
    data changes.

    The OQD_SOURCE_DES EP performs an automatic lookback in the database
    and is guaranteed to return output at the start time of the query if
    the security existed on that day.

    Pass the output of OQD_SOURCE_DES through a JOIN_BY_TIME EP or an
    aggregation EP to create a daily series of descriptive data.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_DES.otq"


class OqdSourceDprcAll(BaseOqdEp):
    """
    OQD_SOURCE_DPRC_ALL

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve a time series of unadjusted
    prices for a symbol for all available pricing exchanges.

    The OQD_SOURCE_DPRC_ALL EP retrieves a time series of all available
    daily unadjusted prices for a symbol.  The OneQuantData exchange code for each
    price series is in the EXCH field.

    To adjust the data for corporate actions, use either the
    OQD_CORP_ACTION EP or the CORP_ACTION EP.

    To retrieve a price series for only one particular exchange, see
    OQD_SOURCE_DPRC_EXCH.  For a price series that tracks either the
    composite series or the primary exchange series, see
    OQD_SOURCE_DPRC_MAIN.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_DPRC_ALL.otq"


class OqdSourceDprcExch(BaseOqdEp):
    """
    OQD_SOURCE_DPRC_EXCH

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve a time series of unadjusted
    prices for a symbol for one particular pricing exchange.

    The OQD_SOURCE_DPRC_ALL EP retrieves a time series of daily unadjusted
    prices for a symbol.  A OneQuantData exchange code must be supplied as
    an argument to the EP.

    To adjust the data for corporate actions, use either the
    OQD_CORP_ACTION EP or the CORP_ACTION EP.

    To retrieve a price series for all exchanges available in OneQuantData, see
    OQD_SOURCE_DPRC_ALL.  For a price series that tracks either the
    composite series or the primary exchange series, see
    OQD_SOURCE_DPRC_MAIN.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters:


    EXCH (string)


    The OneQuantData exchange code for the desired price series.

    The OneQuantData exchange code is formed from the 2 letter ISO country
    code plus the 4 letter ISO MIC code.  For countries that have a
    composite price series, use COMP for the exchange portion of the
    code.  For example, USCOMP is the code for US composite pricing.
    """

    class Parameters:
        exch = "EXCH"
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["exch", "shared_state_variables"]
            return list_val

    __slots__ = ["exch", "_default_exch"]

    def __init__(self, exch="USCOMP", shared_state_variables=""):
        BaseOqdEp.__init__(self, shared_state_variables)
        self._default_exch = "USCOMP"
        self.exch = exch

    def set_exch(self, value):
        self.exch = value
        return self

    def _get_name(self):
        return "OTQ::OQD_SOURCE_DPRC_EXCH.otq"

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.exch != "USCOMP":
            desc += "exch=" + py_to_str(self.exch) + ","
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc


class OqdSourceSho(BaseOqdEp):
    """
    OQD_SOURCE_SHO

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve a time series of shares
    outstanding for a stock.

    The OQD_SOURCE_SHO EP retrieves a time series of shares outstanding
    for a stock.  This EP only applies to stocks or securities that have
    published shares outstanding data.

    The series represents total shares outstanding and is not free float
    adjusted.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_SHO.otq"


class OqdSourceXoid(BaseOqdEp):
    """
    OQD_SOURCE_XOID

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve raw ticks from the OQD_XOID
    database.

    The OQD_XOID database translates internal OneQuantData IDs (OIDs) into
    standard identifiers like CUSIP, SEDOL, or TICKER.

    The OQD_SOURCE_XOID EP will return a consolidated time series of
    symbol translations keyed by exchange code (EXCH) and symbol type (ID_TYPE).  Each
    translation is valid from the timestamp of the tick until the
    timestamp given in field END_DATE.

    This EP performs a lookback to ensure that ticks are available for the
    query start time, if there are valid ticks that cover that time.

    Use the OQD_SOURCE_XOID EP to write custom symbol translation queries.

    Set the tick type for this EP to OQD::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters:


    FIGI_COMPOSITE_FILTER
            (boolean)



    When set to false, all Bloomberg tickers and FIGIs associated with an OID will be returned.
    When set to true, only the composite Bloomberg ticker and FIGI are returned.
    This option does not have an effect on any other symbol types.  The default is false.
    """

    class Parameters:
        figi_composite_filter = "FIGI_COMPOSITE_FILTER"
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["figi_composite_filter", "shared_state_variables"]
            return list_val

    __slots__ = ["figi_composite_filter", "_default_figi_composite_filter"]

    def __init__(self, figi_composite_filter=False, shared_state_variables=""):
        BaseOqdEp.__init__(self, shared_state_variables)
        self._default_figi_composite_filter = False
        self.figi_composite_filter = figi_composite_filter

    def set_figi_composite_filter(self, value):
        self.figi_composite_filter = value
        return self

    def _get_name(self):
        return "OTQ::OQD_SOURCE_XOID.otq"

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.figi_composite_filter:
            desc += "figi_composite_filter=" + py_to_str(self.figi_composite_filter) + ","
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc


class OqdSourceXref(BaseOqdEp):
    """
    OQD_SOURCE_XREF

    &#160;

    Type: Other

    Description: A OneQuantData&#153; source EP to retrieve raw ticks from the XREF database.
    The XREF database contains records used to translate external symbols to the OneQuantData internal identifier (OID).

    The OQD_SOURCE_XREF EP will return a consolidated time series of symbol translations
    keyed by exchange code and symbol type.
    Each translation is valid from the timestamp of the tick until the timestamp given in field END_DATE.
    The symbols for this EP should be of the form SYMBOL^SYMBOL_TYPE
    where SYMBOL is the external symbol and SYMBOL_TYPE is the code for one of of the supported symbol types.

    The currently supported symbol types are TKR, SED, CUS, ISN, and TAQ for ticker,
    SEDOL, CUSIP, ISIN, and TAQ ticker (US only), respectively.

    This EP performs a lookback to ensure that ticks are available for the query start time,
    if there are valid ticks that cover that time.

    Use the OQD_SOURCE_XREF EP to write custom symbol translation queries.

    Set the tick type for this EP to "__OQD__::*".  The correct database and tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_XREF.otq"


class OqdSourceXsym(BaseOqdEp):
    """
    OQD_SOURCE_XSYM

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to retrieve raw ticks from the OQD_XSYM
    database.

    The OQD_XSYM database translates external identifiers like CUSIP,
    SEDOL, or TICKER to internal OneQuantData IDs (OIDs).

    The OQD_SOURCE_XSYM EP will return a consolidated time series of
    symbol translations keyed by exchange code and symbol type.  Each
    translation is valid from the timestamp of the tick until the
    timestamp given in field END_DATE.

    The symbols for this EP should be of the form SYMBOL^SYMBOL_TYPE where SYMBOL is the
    external symbol and SYMBOL_TYPE is the code for one of of the
    supported symbol types.  For example, a symbol IBM^TKR would return a
    set of ticks supplying the OID for stock ticker IBM.

    Other symbol types include CUS, SED, or ISN for CUSIP, SEDOL, or ISIN,
    if those symbol types are included in your OneQuantData subscription.

    Use the OQD_SOURCE_XREF EP to write custom symbol translation queries.

    Set the tick type for this EP to __OQD__::*
    The correct database and
    tick type will be set by the EP.

    Input: None

    Output: A series of ticks.

    Parameters: There are no parameters for this EP.
    """

    def _get_name(self):
        return "OTQ::OQD_SOURCE_XSYM.otq"


class OqdTranslate(BaseOqdEp):
    """
    OQD_TRANSLATE

    &#160;

    Type: Other

    Description: A
    OneQuantData&#153; source EP to perform symbol translation.


    This EP is maintained for backwards compatibility and may be deprecated
    in future releases of OneQuantData.

    The OQD_TRANSLATE EP may be used to write custom symbol translation
    queries.

    For access to the raw underlying data used by the OQD_TRANSLATE EP,
    see the OQD_SOURCE_XSYM and OQD_SOURCE_XOID EPs.

    Symbols supplied to the OQD_TRANSLATE EP should be of the form SYMBOL^SYMBOL_TYPE where SYMBOL is the
    external symbol and SYMBOL_TYPE is the code for one of of the
    supported symbol types.  For example, a symbol IBM^TKR would be used
    in the symbol list to translate the stock ticker IBM into another alternate
    symbology like CUSIP.

    The OQD_TRANSLATE EP does not require that a Onetick reference file is
    loaded or configured.  It may be used in queries where symbol date is
    set to 0.

    See also the REF_DATA and SYMBOLOGY_MAPPING EPs, which require that
    a Onetick reference file is configured and loaded.

    Input: None

    Output: A series of ticks.

    Parameters:


    ASOF_DATE (integer)


    The date on which the translation should apply


    TO_ID (string)


    The symbol type that should be returned.
    The type should be specified as a OneQuantData symbol type abbreviation


    EXCH (string)


    The exchange to which the symbol translation applies.
    The exchange code should be specified as a OneQuantData exchange code
    """

    class Parameters:
        asof_date = "ASOF_DATE"
        to_id = "TO_ID"
        exch = "EXCH"
        shared_state_variables = "SHARED_STATE_VARIABLES"

        @staticmethod
        def list_parameters():
            list_val = ["asof_date", "to_id", "exch", "shared_state_variables"]
            return list_val

    __slots__ = ["asof_date", "_default_asof_date", "to_id", "_default_to_id", "exch", "_default_exch"]

    def __init__(self, asof_date="", to_id="OID", exch="USCOMP", shared_state_variables=""):
        BaseOqdEp.__init__(self, shared_state_variables)
        self._default_asof_date = ""
        self.asof_date = asof_date
        self._default_to_id = "OID"
        self.to_id = to_id
        self._default_exch = "USCOMP"
        self.exch = exch

    def set_asof_date(self, value):
        self.asof_date = value
        return self

    def set_to_id(self, value):
        self.to_id = value
        return self

    def set_exch(self, value):
        self.exch = value
        return self

    def _get_name(self):
        return "OTQ::OQD_TRANSLATE.otq"

    def _to_string(self, ep_name=None, for_repr=False):
        name = ep_name
        if name is None:
            name = self._get_name()
        desc = name + "("
        py_to_str = repr if for_repr else str
        if self.asof_date:
            desc += "asof_date=" + py_to_str(self.asof_date) + ","
        py_to_str = repr if for_repr else str
        if self.to_id != "OID":
            desc += "to_id=" + py_to_str(self.to_id) + ","
        py_to_str = repr if for_repr else str
        if self.exch != "USCOMP":
            desc += "exch=" + py_to_str(self.exch) + ","
        py_to_str = repr if for_repr else str
        if self.shared_state_variables:
            desc += "shared_state_variables=" + py_to_str(self.shared_state_variables) + ","
        desc = desc[:-1]
        if desc != name:
            desc += ")"
        if for_repr:
            return desc + '()' if desc == name else desc
        desc += "\n"
        if len(self._symbols) > 0:
            desc += "symbols=[" + ", ".join(self._symbols) + "]\n"
        if len(self._tick_types) > 0:
            desc += "tick_types=[" + ', '.join(self._tick_types) + "]\n"
        if self._process_node_locally:
            desc += "process_node_locally=True\n"
        if self._node_name:
            desc += "node_name=" + self._node_name + "\n"
        return desc



from typing import TYPE_CHECKING, Optional, Type

import numpy as np

from onetick import py as otp
from onetick.py import types as ott
from onetick.py.compatibility import is_existing_fields_handling_supported
from onetick.py.core._internal._state_objects import _StateColumn
from onetick.py.core.column import _Column, _ColumnAggregation, _LagOperator
from onetick.py.core.column_operations._methods.methods import is_arithmetical, is_compare
from onetick.py.core.column_operations._methods.op_types import are_ints_not_time
from onetick.py.core.column_operations.base import _Operation
from onetick.py.core.cut_builder import _BaseCutBuilder
from onetick.py.core.lambda_object import _LambdaIfElse
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def table(self: 'Source', inplace=False, strict: bool = True, **schema) -> Optional['Source']:
    """
    Set the OneTick and python schemas levels according to the ``schema``
    parameter. The ``schema`` should contain either (field_name -> type) pairs
    or (field_name -> default value) pairs; ``None`` means no specified type, and
    OneTick considers it's as a double type.

    Resulting ticks have the same order as in the ``schema``. If only partial fields
    are specified (i.e. when the ``strict=False``) then fields from the ``schema`` have
    the most left position.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operations should be applied inplace
    strict: bool
        If set to ``False``, all fields present in an input tick will be present in the output tick.
        If ``True``, then only fields specified in the ``schema``.
    schema:
        field_name -> type or field_name -> default value pairs that should be applied on the source.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    | :attr:`Source.schema`
    | :meth:`__getitem__`: the table shortcut
    | **TABLE** OneTick event processor

    Examples
    --------

    Selection case

    >>> data = otp.Ticks(X1=[1, 2, 3],
    ...                  X2=[3, 2, 1],
    ...                  A1=["A", "A", "A"])
    >>> data = data.table(X2=int, A1=str)   # OTdirective: snippet-name: Arrange.set schema;
    >>> otp.run(data)
                         Time  X2 A1
    0 2003-12-01 00:00:00.000   3  A
    1 2003-12-01 00:00:00.001   2  A
    2 2003-12-01 00:00:00.002   1  A

    Defining default values case (note the order)

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data = data.table(Y=0.5, strict=False)
    >>> otp.run(data)
                         Time   Y   X
    0 2003-12-01 00:00:00.000  0.5  1
    1 2003-12-01 00:00:00.001  0.5  2
    2 2003-12-01 00:00:00.002  0.5  3
    """

    def is_time_type_or_nsectime(obj):
        return ott.is_time_type(obj) or isinstance(obj, ott.nsectime)

    def transformer(name, obj):
        if obj is None:
            return name

        res = f'{ott.type2str(ott.get_object_type(obj))} {name}'

        if isinstance(obj, ott._inner_string):
            res = f'{name} {ott.type2str(ott.get_object_type(obj))}'
        if not isinstance(obj, Type) and not is_time_type_or_nsectime(obj):
            res += f' ({ott.value2str(obj)})'
        return res

    def get_type(value):
        if value is None:
            return float
        return ott.get_object_type(value)

    for c_name in list(schema.keys()):
        if self._check_key_is_meta(c_name):
            # meta fields should not be propagated to Table ep
            # otherwise new user-defined field with the same name will appear in schema
            # and using this field will raise an "ambiguous use" error in OneTick
            raise ValueError(f"Can't set meta field {c_name}")

    if not schema:
        return self

    schema_to_set = {c_name: get_type(c_value) for c_name, c_value in schema.items()}

    if strict:
        self.schema.set(**schema_to_set)
    else:
        self.schema.update(**schema_to_set)

    fields = ','.join([transformer(c_name, c_value) for c_name, c_value in schema.items()])

    self.sink(otq.Table(fields=fields, keep_input_fields=not strict))
    for c_name, c_value in schema.items():
        # datetime and nsetime values require onetick built-in functions to be initialized
        # but built-in functions can't be used in table ep so updating columns after the table
        if is_time_type_or_nsectime(c_value):
            self.update({c_name: c_value}, where=self[c_name] == 0, inplace=True)
    self._fix_varstrings()
    return self


def __add_field_parse_value(value):
    if isinstance(value, tuple):
        value, dtype = value
    else:
        dtype = ott.get_object_type(value)

        # pylint: disable-next=unidiomatic-typecheck
        if type(value) is str and len(value) > ott.string.DEFAULT_LENGTH:
            dtype = ott.string[len(value)]

    if issubclass(dtype, bool):
        # according to OneTick transformations
        dtype = float

    if issubclass(dtype, (ott.datetime, ott.date)):
        # according to OneTick transformations
        dtype = ott.nsectime

    # TODO: shouldn't all such logic be in ott.type2str?
    if np.issubdtype(dtype, np.integer):
        dtype = int
    if np.issubdtype(dtype, np.floating):
        dtype = float

    return dtype, value


def _add_field(self: 'Source', key, value):
    if isinstance(value, _ColumnAggregation):
        value.apply(self, key)
        return

    dtype, value = __add_field_parse_value(value)
    type_str = ott.type2str(dtype)
    str_value = ott.value2str(value)

    self.sink(otq.AddField(field=f'{type_str} {key}', value=str_value))

    self.__dict__[key] = _Column(key, dtype, self)


def _update_timestamp(self: 'Source', key, value, str_value):
    if not hasattr(value, "dtype"):
        # A constant value: no need to pre- or post-sort
        self.sink(otq.UpdateField(field=key, value=str_value))
    elif (
        isinstance(value, _Column)
        and not isinstance(value, _StateColumn)
        and hasattr(value, "name")
        and value.name in self.__dict__
    ):
        # An existing, present column: no need to create a temporary one. See PY-253
        need_to_sort = str_value not in ("_START_TIME", "_END_TIME")
        if need_to_sort:
            self.sort(value, inplace=True)
        self.sink(otq.UpdateField(field=key, value=str_value))
        if need_to_sort:
            self.sort(self['Time'], inplace=True)
    elif not is_compare(value) and isinstance(value, (_Operation, _LambdaIfElse)) or is_arithmetical(value):
        # An expression or a statevar: create a temp column with its value, pre- and post- sort.
        self.sink(otq.AddField(field="__TEMP_TIMESTAMP__", value=str_value))
        self.sink(otq.OrderByEp(order_by="__TEMP_TIMESTAMP__ ASC"))
        self.sink(otq.UpdateField(field=key, value="__TEMP_TIMESTAMP__"))
        self.sink(otq.Passthrough(fields="__TEMP_TIMESTAMP__", drop_fields=True))
        self.sort(self['Time'], inplace=True)
    else:
        raise TypeError(f"Illegal type for timestamp assignment: {value.__class__}")


def _replace_positive_lag_operator_with_tmp_column(value):
    """
    Positive lag operator can't be used in UPDATE_FIELD EP,
    so we are replacing it with temporary column.
    """
    if not isinstance(value, _Operation):
        return value, None

    def fun(operation):
        if isinstance(operation, _LagOperator) and operation.index > 0:
            column = operation._op_params[0]
            name = column.name
            if name.startswith("__"):
                raise ValueError(
                    "Column name started with two underscores should be used by system only, "
                    "please do not use such names."
                )
            name = f"__{name}_{operation.index}_NEW__"
            return _Column(name, column.dtype, column.obj_ref, precision=getattr(column, "_precision", None))
        return None

    op, replace_tuples = value._replace_parameters(fun, return_replace_tuples=True)
    return op, {str(new): old for old, new in replace_tuples}


def _update_field(self: 'Source', field, value):

    if isinstance(value, _ColumnAggregation):
        value.apply(self, str(field))
        return

    value, names_mapping = _replace_positive_lag_operator_with_tmp_column(value)
    if names_mapping:
        self.add_fields(names_mapping, inplace=True)

    if isinstance(value, tuple):
        # support to be compatible with adding fields to get rid of some strange problems
        # but really we do not use passed type, because update field does not support it
        value, _ = value

    convert_to_type = None
    str_value = ott.value2str(value)
    value_dtype = ott.get_object_type(value)
    base_type = ott.get_base_type(value_dtype)

    if base_type is bool:
        # according OneTick
        base_type = float

    type_changes = False  # because mantis 0021194
    if base_type is str:
        # update_field non-string field to string field (of any length) or value
        # changes type to default string
        if not issubclass(field.dtype, str):
            field._dtype = str
            type_changes = True

    else:
        if (
            (issubclass(field.dtype, int) or issubclass(field.dtype, float) or issubclass(field.dtype, str))
            and (issubclass(value_dtype, ott.msectime) or issubclass(value_dtype, ott.nsectime))
            and (str(field) != 'TIMESTAMP')
            and (not isinstance(field, _StateColumn))
        ):
            # in OneTick after updating fields with functions that return datetime values
            # the type of column will not change for long and double columns
            # and will change to long (or double in older versions) when updating string column
            # (see BDS-267)
            # That's why we are explicitly setting type for returned value
            convert_to_type = value_dtype
            if issubclass(field.dtype, str):
                # using update_field only for string because update_fields preserves type
                # by default and raises exception if it can't be done
                type_changes = True
        elif issubclass(field.dtype, float) and base_type is int and not isinstance(field, _StateColumn):
            # PY-574 if field was float and int, then
            # it is still float in onetick, so no need to change type on otp level
            convert_to_type = int
        elif (issubclass(field.dtype, ott.msectime) or issubclass(field.dtype, ott.nsectime)) and base_type is int:
            # if field was time type and we add something to it, then
            # no need to change type
            pass
        elif issubclass(field.dtype, str) and base_type is int:
            type_changes = True
            convert_to_type = int
        elif (
            are_ints_not_time(field.dtype, value_dtype) and
            not issubclass(field.dtype, bool) and not issubclass(value_dtype, bool) and
            value_dtype is not field.dtype
            and not isinstance(field, _StateColumn)
        ):
            # case for converting values between int based types, like long or byte
            convert_to_type = value_dtype
        else:
            if issubclass(value_dtype, bool):
                value_dtype = float

            if isinstance(field, _StateColumn):
                pass  # do nothing
            else:
                field._dtype = value_dtype
                type_changes = True

    if isinstance(field, _StateColumn) and (convert_to_type is not None or type_changes):
        raise ValueError(f"The type of the state variable {field.name} can't be changed")

    # for aliases, TIMESTAMP ~ Time as an example
    key = str(field)

    if key == "TIMESTAMP":
        self._update_timestamp(key, value, str_value)

    elif type_changes:
        if value_dtype in [otp.nsectime, int] and issubclass(field.dtype, str):
            # PY-416 string field changing the type from str to datetime leads to losing nanoseconds
            # BE-142 similar issue with int from string: OneTick convert str to float, and then to int
            # so we lose some precision for big integers
            # work around is: make a new column first, delete accessor column
            # and then recreate it with value from temp column
            self.sink(otq.AddField(field=f"_TMP_{key}", value=str_value))
            self.sink(otq.Passthrough(fields=key, drop_fields=True))
            self.sink(otq.AddField(field=f"{key}", value=f"_TMP_{key}"))
            self.sink(otq.Passthrough(fields=f"_TMP_{key}", drop_fields=True))
        else:
            self.sink(otq.UpdateField(field=key, value=str_value))
    else:
        self.sink(otq.UpdateFields(set=key + "=" + str_value))
    if names_mapping:
        self.drop(list(names_mapping), inplace=True)
    if convert_to_type:
        # manual type conversion after update fields for some cases
        self.table(**{key: convert_to_type}, inplace=True, strict=False)


def _validate_before_setting(key, value):
    if key in ["Symbol", "_SYMBOL_NAME"]:
        raise ValueError("Symbol setting is supported during creation only")
    if key == "_state_vars":
        raise ValueError("state field is necessary for keeping state variables and can't be rewritten")
    if isinstance(value, ott.ExpressionDefinedTimeOffset):
        value = value.n
    if isinstance(value, np.generic):
        value = value.item()
    if not (
        ott.is_type_supported(ott.get_object_type(value))
        or isinstance(value, (_Operation, tuple, _ColumnAggregation))
    ):
        raise TypeError(f'It is not allowed to set objects of "{type(value)}" type')
    return value


def __setattr__(self: 'Source', key, value):
    if self._check_key_in_properties(key):
        self.__dict__[key] = value
        return

    # we only allow TIMESTAMP field to be changed
    if self._check_key_is_meta(key) and key not in {'Time', 'TIMESTAMP'}:
        raise ValueError(f"Can't set meta field {key}")

    if isinstance(value, _BaseCutBuilder):
        value(key)
        return

    value = _validate_before_setting(key, value)
    if key in self.__dict__:
        field = self.__dict__[key]
        if issubclass(type(field), _Column):
            self._update_field(field, value)
        else:
            raise AttributeError(f'Column "{key}" not found')
    else:
        assert not (
            isinstance(value, _StateColumn) and value.obj_ref is None
        ), "State variables should be in `state` field"
        self._add_field(key, value)


def __setitem__(self: 'Source', key, value):
    """
    Add new column to the source or update existing one.

    Parameters
    ----------
    key: str
        The name of the new or existing column.
    value: int, str, float, :py:class:`datetime.datetime`, :py:class:`datetime.date`, \
            :py:class:`~onetick.py.Column`, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.string`, \
            :py:class:`otp.date <onetick.py.date>`, :py:class:`otp.datetime <onetick.py.datetime>`, \
            :py:class:`~onetick.py.nsectime`, :py:class:`~onetick.py.msectime`
        The new value of the column.

    See also
    --------
    | **ADD_FIELD** OneTick event processor
    | **UPDATE_FIELD** OneTick event processor

    Examples
    --------
    >>> data = otp.Tick(A='A')
    >>> data['D'] = otp.datetime(2022, 2, 2)
    >>> data['X'] = 1
    >>> data['Y'] = data['X']
    >>> data['X'] = 12345
    >>> data['Z'] = data['Y'].astype(str) + 'abc'
    >>> otp.run(data)
            Time  A          D      X  Y     Z
    0 2003-12-01  A 2022-02-02  12345  1  1abc
    """

    return self.__setattr__(key, value)


@inplace_operation
def add_fields(self: 'Source', fields: dict, override: bool = False, inplace=False):
    """
    Add new columns to the source.

    Parameters
    ----------
    fields: dict
        The dictionary of the names of the new fields and their values.
        The types of supported values in the dictionary are the same as in :meth:`Source.__setitem__`.
    override: bool
        If *False* then exception will be raised for existing fields.
        If *True* then exception will not be raised and the field will be overridden.
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``.

    See also
    --------
    | :meth:`Source.__setitem__`
    | **ADD_FIELDS** OneTick event processor

    Examples
    --------

    Add new fields specified in the dictionary:

    >>> data = otp.Tick(A=1)
    >>> data = data.add_fields({
    ...     'D': otp.datetime(2022, 2, 2),
    ...     'X': 12345,
    ...     'Y': data['A'],
    ...     'Z': data['A'].astype(str) + 'abc',
    ... })
    >>> otp.run(data)
            Time  A           D      X  Y     Z
    0 2003-12-01  1  2022-02-02  12345  1  1abc

    Parameter ``override`` can be used to rewrite existing fields:

    .. testcode::
        :skipif: not is_existing_fields_handling_supported()

        data = otp.Tick(A=1)
        data = data.add_fields({'A': 2, 'B': 'b'}, override=True)
        df = otp.run(data)
        print(df)

    .. testoutput::

                Time  A  B
        0 2003-12-01  2  b
    """
    fields_parsed = {}

    kwargs = {}
    if override:
        if not is_existing_fields_handling_supported():
            raise ValueError("Parameter 'override' is not supported on this OneTick build")
        kwargs['existing_fields_handling'] = 'OVERRIDE'

    for key, value in fields.items():
        if self._check_key_in_properties(key):
            raise ValueError("Class properties can't be set with add_fields method")
        if self._check_key_is_meta(key):
            raise ValueError(f"Can't set meta field {key}")
        if not override and key in self.schema:
            raise ValueError(f"Field '{key}' is already in schema")

        # TODO: _validate_before_setting and __add_field_parse_value both modify value, need to refactor
        value = _validate_before_setting(key, value)
        dtype, value = __add_field_parse_value(value)
        fields_parsed[key] = (_Column(key, dtype, self), f'{ott.type2str(dtype)} {key}={ott.value2str(value)}')

    fields_str = ','.join(field_str for _, field_str in fields_parsed.values())
    self.sink(otq.AddFields(fields=fields_str, **kwargs))

    for key, (column, _) in fields_parsed.items():
        self.__dict__[key] = column

    return self


@inplace_operation
def update(self: 'Source', if_set, else_set=None, where=1, inplace=False) -> 'Source':
    """
    Update field of the Source or state variable.

    Parameters
    ----------
    if_set: dict
        Dictionary <field name>: <expression>.
    else_set: dict, optional
        Dictionary <field name>: <expression>
    where: expression, optional
        Condition of updating.

        If ``where`` is True the fields from ``if_set`` will be updated with corresponding expression.

        If ``where`` is False, the fields from ``else_set`` will be updated with corresponding expression.

    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``.

    See also
    --------
    **UPDATE_FIELD** and **UPDATE_FIELDS** OneTick event processors

    Examples
    --------

    Columns can be updated with this method:

    >>> # OTdirective: snippet-name: Arrange.conditional update;
    >>> t = otp.Ticks({'X': [1, 2, 3],
    ...                'Y': [4, 5, 6],
    ...                'Z': [1, 0, 1]})
    >>> t = t.update(if_set={'X': t['X'] + t['Y']},
    ...              else_set={'X': t['X'] - t['Y']},
    ...              where=t['Z'] == 1)
    >>> otp.run(t) # OTdirective: snippet-example;
                         Time  X  Y  Z
    0 2003-12-01 00:00:00.000  5  4  1
    1 2003-12-01 00:00:00.001 -3  5  0
    2 2003-12-01 00:00:00.002  9  6  1

    State variables can be updated too:

    >>> t = otp.Ticks({'X': [1, 2, 3],
    ...                'Y': [4, 5, 6],
    ...                'Z': [1, 0, 1]})
    >>> t.state_vars['X'] = 0
    >>> t = t.update(if_set={t.state_vars['X']: t['X'] + t['Y']},
    ...              else_set={t.state_vars['X']: t['X'] - t['Y']},
    ...              where=t['Z'] == 1)
    >>> t['UX'] = t.state_vars['X']
    >>> otp.run(t)
                         Time  X  Y  Z  UX
    0 2003-12-01 00:00:00.000  1  4  1   5
    1 2003-12-01 00:00:00.001  2  5  0  -3
    2 2003-12-01 00:00:00.002  3  6  1   9
    """
    if else_set is None:
        else_set = {}

    if len(if_set) == 0 or not isinstance(if_set, dict):
        raise ValueError(f"'if_set' parameter should be non empty dict, but got '{if_set}' of type '{type(if_set)}'")

    def _prepare(to_prepare):
        result = {}

        for in_obj, out_obj in to_prepare.items():
            if isinstance(in_obj, _StateColumn):
                result[in_obj] = out_obj
            elif isinstance(in_obj, _Column):
                result[in_obj.name] = out_obj
            elif isinstance(in_obj, str):
                result[in_obj.strip()] = out_obj
            else:
                raise AttributeError(f"It is not supported to update item '{in_obj}' of type '{type(in_obj)}'")

        return result

    def _validate(to_validate):
        result = {}
        for in_key, out_obj in to_validate.items():
            if isinstance(in_key, _StateColumn):
                in_dtype = in_key.dtype
            else:
                if not (in_key in self.__dict__ and isinstance(self.__dict__[in_key], _Column)):
                    raise AttributeError(f"There is no '{in_key}' column to update")

                if self._check_key_is_meta(in_key):
                    raise ValueError(f"Can't set meta field {in_key}")

                in_dtype = self.schema[in_key]

            dtype = ott.get_object_type(out_obj)
            if not ott.is_type_supported(dtype):
                raise TypeError(f"Deduced type of object {out_obj} is not supported: {dtype}")
            # will raise exception if types are incompatible
            _ = ott.get_type_by_objects([dtype, in_dtype])

            if isinstance(in_key, _StateColumn):
                in_key = str(in_key)

            if isinstance(out_obj, bool):
                out_obj = int(out_obj)

            result[in_key] = out_obj

        return result

    # prepare and validate
    items = _validate(_prepare(if_set))
    else_items = _validate(_prepare(else_set))

    if isinstance(where, bool):
        where = int(where)

    if not (getattr(where, "dtype", None) is bool or isinstance(where, int)):
        raise ValueError(f"Where has not supported type '{type(where)}'")

    # apply
    set_rules = [f"{key}=({ott.value2str(value)})" for key, value in items.items()]
    else_set_rules = [f"{key}=({ott.value2str(value)})" for key, value in else_items.items()]

    self.sink(otq.UpdateFields(set=",".join(set_rules), else_set=",".join(else_set_rules), where=str(where)))
    return self



from .functions import (compute, max, min, high_tick, low_tick, high_time, low_time, first, last, first_time, last_time,
                        count, vwap, first_tick, last_tick, distinct, sum, average, mean, stddev, tw_average, median,
                        ob_num_levels, ob_size, ob_snapshot, ob_snapshot_wide, ob_snapshot_flat, ob_summary, ob_vwap,
                        generic, correlation, option_price, ranking, variance, percentile, find_value_for_percentile,
                        exp_w_average, exp_tw_average, standardized_moment, portfolio_price, multi_portfolio_price,
                        return_ep, implied_vol, linear_regression, partition_evenly_into_groups)

try:
    from .num_distinct import num_distinct
except ImportError:
    pass



import os
import re
import warnings
import shutil
from datetime import datetime
from typing import List

from locator_parser.io import FileReader, FileWriter, PrintWriter
from locator_parser.actions import Add, GetAll, Delete
from locator_parser.common import apply_actions
from locator_parser import locator as _locator
from locator_parser import acl as _acl
from abc import ABC, abstractmethod
from onetick.py.otq import otli
from . import utils
from . import license as _license
from . import db as _db
from . import servers as _servers
from . import configuration
from .db._inspection import databases as _databases


class EntityOperationFailed(Exception):
    """
    Raise when operation with an entity for a config
    in this module failed
    """


class MultipleSessionsException(Exception):
    """
    Raises when user tries to initiate a new one Session
    when another one is already used
    """


def _check_entities(entities):
    if not entities:
        raise ValueError("At least one argument in parameter 'entities' is expected.")


def _apply_to_entities(cfg, operations):
    """
    Function generalizes operations for locators and ACLs:
    tries to apply ``operations``, and does rollback in case
    OneTick config is invalid.
    """
    if not operations:
        return

    for entities, func, roll_back_func in operations:
        _check_entities(entities)
        result = func(entities)
        if not result:
            entity_name = entities[0].__class__.__name__
            entities = list(map(str, entities))
            raise EntityOperationFailed(
                f'Operation {func.__name__} for {entity_name}s {entities}'
                f' for {cfg.__class__.__name__} "{cfg.path}" has failed'
            )
    try:
        cfg.reload()
    except Exception:
        for entities, func, roll_back_func in reversed(operations):
            roll_back_func(entities)
        raise


class _FileHandler(ABC):
    def __init__(self, file_h=None, clean_up=utils.default, copied=True, session_ref=None):
        self._file = file_h
        self._clean_up = clean_up
        # flag to understand whether we work with externally passed files;
        # it is set and affects logic, when copy=False
        self._copied = copied
        self._session_ref = session_ref

    @property
    def path(self):
        return self._file.path

    @property
    def file(self):
        return self._file

    def copy(self, clean_up=utils.default, copy=True, session_ref=None):
        return self.__class__(self.path, clean_up=clean_up, copy=copy, session_ref=session_ref)  # pylint: disable=E1123

    @abstractmethod
    def cleanup(self):
        pass

    @staticmethod
    def _db_in_dbs_case_insensitive(db_id: str, databases: List[str]):
        for db_name in databases:
            if db_id.upper() == db_name.upper():
                return True
        return False


class _CommonBuilder(ABC):
    def __init__(self, src=None, clean_up=utils.default, copy=True, session_ref=None):
        self.src = src
        self.clean_up = clean_up
        self.copy = copy
        self.session_ref = session_ref

    @abstractmethod
    def build(self):
        pass


class ACL(_FileHandler):
    class User(str):
        """
        Subclass represents an ACL user
        """

        pass

    def __init__(self, path=None, clean_up=utils.default, copy=True, session_ref=None):
        """
        Class representing OneTick database access list file.
        ACL is the file that describes the list of the users
        that are allowed to access the database and what permissions do they have.

        Parameters
        ----------
        path: str
            A path to custom acl file. Default is `None`, that means to generate a temporary acl file.
        clean_up: bool
            If `True`, then temporary acl file will be removed when ACL object will be destroyed. It is
            helpful for debug purpose.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
        copy: bool
            If `True`, then the passed custom acl file by the ``path`` parameter will be copied first before
            usage. It might be used when you want to work with a custom acl file, but don't want to change
            the original file; in that case a custom acl file will be copied into a temporary file and
            every request for modification will be executed for that temporary file. Default is `True`.
        """

        copied = True

        # TODO: implement this logic later
        # if copy is None and path is not None:
        #    # if copy rule is not specified, but path is specified
        #    # then we set copy to True with safety goal, otherwise
        #    # we would might change a permanent file without user
        #    # acknowledgment
        #    copy = True
        #    raise Warning("You set the ACL file, but have not specify a copy rule. "
        #                  "We copy it with safety goal, and it means you will work "
        #                  "with copied file instead of passed. If you want to work "
        #                  "with passed file directly, then you could set the 'copy' "
        #                  "parameter to True.")

        # if path is set, then copy file, we should not work directly
        # with externally passed files
        if copy:
            if path:
                file_h = utils.TmpFile(suffix=".acl", clean_up=clean_up)
                shutil.copyfile(path, file_h.path)
            else:
                file_h = utils.tmp_acl(clean_up=clean_up)
        else:
            if path:
                file_h = utils.PermanentFile(path)
                copied = False
            else:
                file_h = utils.tmp_acl(clean_up=clean_up)

        assert file_h is not None

        super().__init__(file_h, clean_up=clean_up, copied=copied, session_ref=session_ref)

        self._added_dbs = []

    def cleanup(self):
        self._remove_db(self._added_dbs)
        self._added_dbs = []
        self.reload()

    def _apply_actions(self, actions, print_writer=False):
        writer = PrintWriter() if print_writer else FileWriter(self.path)
        flush = False if print_writer else True
        return apply_actions(_acl.parse_acl, FileReader(self.path), writer, actions, flush=flush)

    def _add_db(self, dbs):
        actions = []
        for db in dbs:
            actions.append(Add(_acl.DB(id=db.id, read_access="true")))

            permissions = {}
            if db._write:
                permissions["write_access"] = "true"
            if hasattr(db, "_destroy_access") and db._destroy_access:
                permissions["destroy_access"] = "true"
            if db._minimum_start_date:
                permissions['minimum_start_date'] = db._minimum_start_date.strftime('%Y%m%d')
            if db._maximum_end_date:
                permissions['maximum_end_date'] = db._maximum_end_date.strftime('%Y%m%d')

            if permissions:
                action = Add(_acl.Allow(role="Admin", **permissions))
                action.add_where(_acl.DB, id=db.id)
                actions.append(action)

        return self._apply_actions(actions)

    def _remove_db(self, dbs):
        actions = []
        for db in dbs:
            action = Delete()
            action.add_where(_acl.DB, id=db.id)
            actions.append(action)

        return self._apply_actions(actions)

    def _add_user(self, users):
        actions = []
        for user in users:
            action = Add(_acl.User(name=user))
            action.add_where(_acl.Role, name="Admin")
            actions.append(action)

        return self._apply_actions(actions)

    def _remove_user(self, users):
        actions = []
        for user in users:
            action = Delete()
            action.add_where(_acl.Role, name="Admin")
            action.add_where(_acl.User, name=user)
            actions.append(action)

        return self._apply_actions(actions)

    def add(self, *entities):
        """
        Add entities to the ACL and reload it.
        If it fails, then tries to roll back to the original state.

        Parameters
        ----------
        entities: DB or ACL.User

        Raises
        ------
        TypeError, EntityOperationFailed
        """
        _check_entities(entities)

        dbs = []
        users = []
        for entity in entities:
            if isinstance(entity, _db.DB):
                if self._db_in_dbs_case_insensitive(entity.id, self.databases):
                    if '//' not in entity.name:
                        warnings.warn(f"Database '{entity.id}' is already added to the ACL"
                                      " and will not be rewritten with this command."
                                      f" Notice that databases' names are case insensitive.",
                                      stacklevel=2)
                    continue
                dbs.append(entity)
            elif isinstance(entity, ACL.User):
                users.append(entity)
            else:
                raise TypeError(f'Entity of type "{type(entity)}" is not supported')

        operations = []
        if dbs:
            operations.append((dbs, self._add_db, self._remove_db))
        if users:
            operations.append((users, self._add_user, self._remove_user))

        _apply_to_entities(self, operations)

        self._added_dbs.extend(dbs)

    def remove(self, *entities):
        """
        Remove entities from the ACL and reload it.
        If it fails, then tries to roll back to the original state.

        Parameters
        ----------
        entities: DB or ACL.User

        Raises
        ------
        ValueError, TypeError, EntityOperationFailed
        """
        _check_entities(entities)

        dbs = []
        users = []
        for entity in entities:
            if isinstance(entity, _db.DB):
                if entity not in self._added_dbs:
                    raise ValueError(f'DB "{entity}" was not added')
                dbs.append(entity)
            elif isinstance(entity, ACL.User):
                users.append(entity)
            else:
                raise TypeError(f'Entity of type "{type(entity)}" is not supported')

        operations = []
        if dbs:
            operations.append((dbs, self._remove_db, self._add_db))
        if users:
            operations.append((users, self._remove_user, self._add_user))

        _apply_to_entities(self, operations)

        for db in dbs:
            self._added_dbs.remove(db)

    def reload(self, db=None):
        if self._session_ref is not None:
            return utils.reload_config(db, config_type='ACCESS_LIST')
        return None

    def _read_dbs(self):
        get_db = GetAll()
        get_db.add_where(_acl.DB)
        self._apply_actions([get_db], print_writer=True)
        return list(map(lambda x: x.id, get_db.result))

    def _dbs(self):
        action = GetAll()
        action.add_where(_acl.DB)
        self._apply_actions([action], print_writer=True)
        return list(map(lambda x: x.id, action.result))

    def _users(self):
        action = GetAll()
        action.add_where(_acl.Role, name="Admin")
        action.add_where(_acl.User)
        self._apply_actions([action], print_writer=True)
        return list(map(lambda x: x.name, action.result))

    @property
    def databases(self):
        return self._dbs()

    @property
    def users(self):
        return self._users()


class ACLBuilder(_CommonBuilder):
    def build(self):
        params = {"clean_up": self.clean_up, "copy": self.copy, "session_ref": self.session_ref}

        if isinstance(self.src, str):
            return ACL(self.src, **params)
        elif isinstance(self.src, utils.File):
            return ACL(self.src.path, **params)
        elif isinstance(self.src, ACL):
            return self.src.copy(**params)
        elif self.src is None:
            return ACL(**params)

        raise ValueError(f'It is not allowed to build ACL from the object of type "{type(self.src)}"')


class Locator(_FileHandler):
    def __init__(self, path=None, clean_up=utils.default, copy=True, empty=False, session_ref=None):
        """
        Class representing OneTick database locator.
        Locator is the file that describes database name, location and other options.

        Parameters
        ----------
        path: str
            A path to custom locator file. Default is `None`, that means to generate a temporary locator.
        clean_up: bool
            If True, then temporary locator will be removed when Locator object will be destroyed. It is
            helpful for debug purpose.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
        copy: bool
            If `True`, then the passed custom locator by the ``path`` parameter will be copied firstly before
            usage. It might be used when you want to work with a custom locator, but don't want to change
            the original file; in that case a custom locator will be copied into a temporary locator and
            every request for modification will be executed for that temporary locator. Default is `True`.
        empty: bool
            If `True`, then a temporary locator will have no databases, otherwise it will have default
            otp.config.default_db and COMMON databases. Default is `False`.
        """
        copied = True

        # if path is set, then copy file, we should not work directly
        # with externally passed files
        if copy:
            if path:
                file_h = utils.TmpFile(".locator", clean_up=clean_up)
                shutil.copyfile(path, file_h)
            else:
                file_h = utils.tmp_locator(clean_up=clean_up, empty=empty)

        else:
            if path:
                file_h = utils.PermanentFile(path)
                copied = False
            else:
                file_h = utils.tmp_locator(clean_up=clean_up, empty=empty)

        assert file_h is not None

        super().__init__(file_h, clean_up=clean_up, copied=copied, session_ref=session_ref)

        self._added_dbs = []
        self._added_ts = []

    def cleanup(self):
        self._remove_db(self._added_dbs)
        self._remove_ts(str(server) for server in self._added_ts)
        self._added_dbs = []
        self._added_ts = []
        self.reload()

    @property
    def databases(self):
        return self._dbs()

    @property
    def tick_servers(self):
        return self._ts()

    def reload(self, db_=None):
        if self._session_ref is not None:
            return utils.reload_config(db_, config_type='LOCATOR')
        return None

    def _apply_actions(self, actions, print_writer=False):
        writer = PrintWriter() if print_writer else FileWriter(self.path)
        flush = False if print_writer else True
        return apply_actions(_locator.parse_locator, FileReader(self.path), writer, actions, flush=flush)

    def _dbs(self):
        action = GetAll()
        action.add_where(_locator.DB)
        self._apply_actions([action], print_writer=True)
        return list(map(lambda x: x.id, action.result))

    def _ts(self):
        get_ts = GetAll()
        get_ts.add_where(_locator.TickServers)
        get_ts.add_where(_locator.ServerLocation)
        self._apply_actions([get_ts], print_writer=True)
        return [location.location for location in get_ts.result]

    def _add_db(self, dbs):
        actions = []
        for db in dbs:
            actions.append(Add(_locator.DB(id=db.id, **db.properties)))
            for location in db.locations:
                action = Add(_locator.Location(**location))
                action.add_where(_locator.DB, id=db.id)
                actions.append(action)

            for raw_db in db.raw_data:
                common = {k: v for k, v in raw_db.items() if k not in {'id', 'locations'}}
                action = Add(_locator.RawDB(id=raw_db['id'], **common))
                action.add_where(_locator.DB, id=db.id)
                actions.append(action)
                for location in raw_db['locations']:
                    action = Add(_locator.Location(**location))
                    action.add_where(_locator.DB, id=db.id)
                    action.add_where(_locator.RawDB, id=raw_db['id'])
                    actions.append(action)

            if db.feed:
                options = {k: v for k, v in db.feed.items() if k != 'type'}
                action = Add(_locator.Feed(type=db.feed['type']))
                action.add_where(_locator.DB, id=db.id)
                actions.append(action)
                action = Add(_locator.FeedOptions(**options))
                action.add_where(_locator.DB, id=db.id)
                action.add_where(_locator.Feed, type=db.feed['type'])
                actions.append(action)

        return self._apply_actions(actions)

    def _remove_db(self, dbs):
        actions = []
        for db in dbs:
            action = Delete()
            action.add_where(_locator.DB, id=db.id)
            actions.append(action)

        return self._apply_actions(actions)

    def _add_ts(self, servers):
        """
        Add servers to locator file (without reloading)

        Parameters
        ----------
        servers: RemoteTS
            Servers to be added to locator.
        """
        actions = []
        for server in servers:
            if server.cep:
                actions.append(Add(_locator.CEPServerLocation(location=str(server))))
            else:
                actions.append(Add(_locator.ServerLocation(location=str(server))))

        return self._apply_actions(actions)

    def _remove_ts(self, servers):
        """
        Remove servers from locator file (without reloading)

        Parameters
        ----------
        servers: RemoteTS
            Servers to remove from locator
        """
        actions = []
        for server in servers:
            action = Delete()
            action.add_where(_locator.ServerLocation, location=str(server))
            actions.append(action)

        return self._apply_actions(actions)

    def _add_locator(self, locators):
        """
        Add references to locators

        Parameters
        ----------
        locators: Locator
        """
        actions = []
        for locator in locators:
            actions.append(Add(_locator.Include(path=locator.path)))

        return self._apply_actions(actions)

    def _remove_locator(self, locators):
        """
        Remove references for locators

        Parameters
        ----------
        locators: Locator
        """
        actions = []
        for locator in locators:
            action = Delete()
            action.add_where(_locator.Include, path=locator.path)
            actions.append(action)

        return self._apply_actions(actions)

    def add(self, *entities):
        """
        Add entities to the locator and reload it.
        If it fails, then tries to roll back to the original state.

        Parameters
        ----------
        entities: DB, RemoteTS or Locator

        Raises
        ------
        TypeError, EntityOperationFailed
        """
        _check_entities(entities)

        dbs = []
        servers = []
        locators = []
        for entity in entities:
            if isinstance(entity, _db.db._DB):
                if self._db_in_dbs_case_insensitive(entity.id, self.databases):
                    if '//' not in entity.name:
                        warnings.warn(f"Database '{entity.id}' is already added to the Locator"
                                      " and will not be rewritten with this command."
                                      f" Notice that databases' names are case insensitive.", stacklevel=2)
                    continue
                dbs.append(entity)
            elif isinstance(entity, _servers.RemoteTS):
                servers.append(entity)
            elif isinstance(entity, Locator):
                locators.append(entity)
            else:
                raise TypeError(f'Entity of type "{type(entity)}" is not supported')

        operations = []
        if dbs:
            operations.append((dbs, self._add_db, self._remove_db))
        if servers:
            operations.append((servers, self._add_ts, self._remove_ts))
        if locators:
            operations.append((locators, self._add_locator, self._remove_locator))

        _apply_to_entities(self, operations)

        self._added_dbs.extend(dbs)
        self._added_ts.extend(servers)

    def remove(self, *entities):
        """
        Remove entities from the locator and reload it.
        If it fails, then tries to roll back to the original state.

        Raises
        ------
        ValueError, TypeError, EntityOperationFailed
        """
        _check_entities(entities)

        dbs = []
        servers = []
        locators = []
        for entity in entities:
            if isinstance(entity, _db.db._DB):
                if entity not in self._added_dbs:
                    raise ValueError(f'DB "{entity}" was not added')
                dbs.append(entity)
            elif isinstance(entity, _servers.RemoteTS):
                if entity not in self._added_ts:
                    raise ValueError(f'Tick server "{entity}" was not added')
                servers.append(entity)
            elif isinstance(entity, Locator):
                locators.append(entity)
            else:
                raise TypeError(f'Entity of type "{type(entity)}" is not supported')

        operations = []
        if dbs:
            operations.append((dbs, self._remove_db, self._add_db))
        if servers:
            operations.append((servers, self._remove_ts, self._add_ts))
        if locators:
            operations.append((locators, self._remove_locator, self._add_locator))

        _apply_to_entities(self, operations)

        for db in dbs:
            self._added_dbs.remove(db)
        for server in servers:
            self._added_ts.remove(server)

    def __contains__(self, item):
        if str(item) in self.databases:
            return True
        return False


class LocatorBuilder(_CommonBuilder):
    def build(self):
        params = {"clean_up": self.clean_up, "copy": self.copy, "session_ref": self.session_ref}

        if isinstance(self.src, str):
            return Locator(self.src, **params)
        elif isinstance(self.src, utils.File):
            return Locator(self.src.path, **params)
        elif isinstance(self.src, Locator):
            return self.src.copy(**params)
        elif isinstance(self.src, _servers.RemoteTS):
            locator = Locator(empty=True, **params)
            locator.add(self.src)
            return locator
        elif self.src is None:
            return Locator(**params)

        raise ValueError(f'It is not allowed to build Locator from the object of type "{type(self.src)}"')


class Config(_FileHandler):

    _CONFIG_VARIABLES_PASSED_VIA_THEIR_OWN_PARAMETER = {
        'ACCESS_CONTROL_FILE': 'acl',
        'DB_LOCATOR.DEFAULT': 'locator',
        'OTQ_FILE_PATH': 'otq_path',
        'CSV_FILE_PATH': 'csv_path',
        'LICENSE_REPOSITORY_DIR': 'license',
        'ONE_TICK_LICENSE_FILE': 'license',
    }

    def __init__(
        self,
        config=None,
        locator=None,
        acl=None,
        otq_path=None,
        csv_path=None,
        clean_up=utils.default,
        copy=True,
        session_ref=None,
        license=None,
        variables=None,
    ):
        """
        Parameters
        ----------
        config: path or Config
            Allows to specify a custom config. None is to use temporary generated config. Default is None.
        locator: Locator
            Allows to specify a custom locator file. None is to use temporary generated locator. Default is None.
        acl: ACL
            Allows to specify a custom acl file. None is to use temporary generated acl. Default is None.
        otq_path: list of paths to lookup queries
            OTQ_PATH parameter in the OneTick config file. Default is None, that is equal to the empty list.
        csv_path: list of paths to lookup csv files
            CSV_PATH parameter in the OneTick config file. Default is None, that is equal to the empty list.
        clean_up: bool
            If True, then temporary config file will be removed when the Config instance will be destroyed.
            It is helpful for debug purpose.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
        copy: bool
            If True, then the passed custom config file will be copied firstly before any usage with it.
            It might be used when you want to work with a custom config file, but don't want to change to
            change the original file; in that case a custom config will be copied into a temporary config
            file and every request for modification will be executed for that temporary config. Default
            is True.
        license: instance from the onetick.py.license module
            License to use. If it is not set, then onetick.py.license.Default is used.
        variables: dict
            Other values to pass to config.
        """
        if config and (locator or acl):
            raise ValueError("It is not allowed to use 'config' parameter along with 'locator' or 'acl'")

        # builders that construct locator and acl based on parameters
        acl_builder = ACLBuilder(src=acl, clean_up=clean_up, copy=copy, session_ref=session_ref)
        locator_builder = LocatorBuilder(src=locator, clean_up=clean_up, copy=copy, session_ref=session_ref)
        config_copied = True

        if config:
            # copy passed file, we should not work with externally passed files
            if copy and not os.getenv('OTP_WEBAPI_TEST_MODE'):
                self._file = utils.TmpFile(".cfg", clean_up=clean_up)
                config_path = config.path if isinstance(config, Config) else config

                shutil.copyfile(config_path, self._file.path)
            else:
                self._file = utils.PermanentFile(config)
                config_copied = False

            if utils.is_param_in_config(self._file.path, "ACCESS_CONTROL_FILE"):
                acl_builder.src = utils.get_config_param(self._file.path, "ACCESS_CONTROL_FILE")

            if utils.is_param_in_config(self._file.path, "DB_LOCATOR.DEFAULT"):
                locator_builder.src = utils.get_config_param(self._file.path, "DB_LOCATOR.DEFAULT")

        else:
            self._file = utils.tmp_config(clean_up=clean_up)

        self._acl = acl_builder.build()
        # it is used in onetick-py-test
        os.environ["ONE_TICK_SESSION_ACL_PATH"] = self._acl.path
        self._locator = locator_builder.build()
        # it is used in onetick-py-test
        os.environ["ONE_TICK_SESSION_LOCATOR_PATH"] = self._locator.path

        super().__init__(self._file, clean_up=clean_up, copied=config_copied, session_ref=session_ref)

        # Here we can start to modify files - they are either copied or generated
        # ------------------------------------------------------------------------

        utils.modify_config_param(self.path, "ACCESS_CONTROL_FILE", self._acl.path, throw_on_missing=False)
        utils.modify_config_param(self.path, "DB_LOCATOR.DEFAULT", self._locator.path, throw_on_missing=False)

        if otq_path:
            otq_path = map(str, otq_path)
            utils.modify_config_param(self.path, "OTQ_FILE_PATH", ",".join(otq_path), throw_on_missing=False)
        if csv_path:
            csv_path = map(str, csv_path)
            utils.modify_config_param(self.path, "CSV_FILE_PATH", ",".join(csv_path), throw_on_missing=False)
        variables = variables or {}
        for parameter_name, parameter_value in variables.items():
            if parameter_name in self._CONFIG_VARIABLES_PASSED_VIA_THEIR_OWN_PARAMETER:
                raise ValueError(f'Variable {parameter_name} should be set via '
                                 f'{self._CONFIG_VARIABLES_PASSED_VIA_THEIR_OWN_PARAMETER[parameter_name]} parameter')
            if isinstance(parameter_value, list):
                parameter_value = ",".join(map(str, parameter_value))
            utils.modify_config_param(self.path, parameter_name, parameter_value, throw_on_missing=False)

        # set license
        # ---------------------------
        custom_license = utils.is_param_in_config(self.path, "LICENSE_REPOSITORY_DIR")
        custom_license &= utils.is_param_in_config(self.path, "ONE_TICK_LICENSE_FILE")
        if license:
            self._license = license
        else:
            if custom_license:
                lic_file = utils.get_config_param(self.path, "ONE_TICK_LICENSE_FILE")
                lic_dir = utils.get_config_param(self.path, "LICENSE_REPOSITORY_DIR")
                self._license = _license.Custom(lic_file, lic_dir)
            else:
                if isinstance(locator, _servers.RemoteTS):
                    self._license = _license.Remote()
                else:
                    self._license = _license.Default()

        if not custom_license:  # no need to set already defined custom values
            if self._license.dir:
                utils.modify_config_param(self.path,
                                          "LICENSE_REPOSITORY_DIR",
                                          self._license.dir,
                                          throw_on_missing=False)
            if self._license.file:
                utils.modify_config_param(self.path,
                                          "ONE_TICK_LICENSE_FILE",
                                          self._license.file,
                                          throw_on_missing=False)

    @property
    def acl(self):
        return self._acl

    @property
    def locator(self):
        return self._locator

    @property
    def license(self):
        return self._license

    def copy(self, clean_up=utils.default, copy=True, session_ref=None):
        """ overridden version of copy """
        return self.__class__(self.path, clean_up=clean_up, copy=copy, session_ref=session_ref, license=self._license)

    @staticmethod
    def build(obj=None, clean_up=utils.default, copy=True, session_ref=None):
        params = {"clean_up": clean_up, "copy": copy, "session_ref": session_ref}

        if isinstance(obj, str):
            return Config(obj, **params)
        elif isinstance(obj, utils.File):
            return Config(obj.path, **params)
        elif isinstance(obj, Config):
            return obj.copy(**params)
        elif obj is None:
            return Config(**params)

        raise ValueError(f'It is not allowed to build Config from the object of type "{type(obj)}"')

    def cleanup(self):
        # no logic to clean up content

        self._acl.cleanup()
        self._locator.cleanup()

    @property
    def otq_path(self):
        try:
            return utils.get_config_param(self.path, "OTQ_FILE_PATH")
        except AttributeError:
            return None


class PerformanceMetricsParser:
    metrics_fields = {
        "user_time": float,
        "system_time": float,
        "elapsed_time": float,
        "virtual_memory": int,
        "virtual_memory_peak": int,
        "working_set": int,
        "working_set_peak": int,
        "disk_read": int,
        "disk_write": int,
    }

    metrics_units = {
        "user_time": "s",
        "system_time": "s",
        "elapsed_time": "s",
        "virtual_memory": "bytes",
        "virtual_memory_peak": "bytes",
        "working_set": "bytes",
        "working_set_peak": "bytes",
        "disk_read": "bytes",
        "disk_write": "bytes",
    }

    def __init__(self):
        self._metrics = {}

    def parse(self, log_file: str):
        with open(log_file, "r") as log_file_io:
            for log_line in log_file_io:
                result = re.search(r"Performance Metrics FINAL ([\w\s]+) \((\w+)\): (.+)$", log_line.strip())
                if not result:
                    continue

                metric_name, metric_units, metric_value = result.groups()
                metric_key = "_".join(metric_name.lower().split())

                if metric_key not in self.metrics_fields:
                    raise ValueError(
                        f"Unexpected performance metric `{metric_key}` was found in the log: \"{log_line.strip()}\""
                    )

                if metric_units != self.metrics_units[metric_key]:
                    raise ValueError(
                        f"Unexpected units for metric `{metric_key}`: "
                        f"expected `{self.metrics_units[metric_key]}`, got `{metric_units}`."
                    )

                if metric_key in self._metrics:
                    raise KeyError(f"Metric `{metric_key}` was already saved, but another entry was found in the log.")

                try:
                    metric_value = self.metrics_fields[metric_key](metric_value)
                except Exception as exc:
                    raise ValueError(
                        f"Failed to parse and convert metric {metric_key} "
                        f"from `str` to `{self.metrics_fields[metric_key].__name__}`: \"{metric_value}\""
                    ) from exc

                self._metrics[metric_key] = {
                    "name": metric_name,
                    "value": metric_value,
                    "units": self.metrics_units[metric_key],
                }

        not_found_keys = set(self.metrics_fields.keys()).difference(self._metrics.keys())
        if not_found_keys:
            raise RuntimeError("Expected performance metrics not found in the log: " + ", ".join(not_found_keys))

    @property
    def metrics(self):
        return self._metrics.copy()


class Session:
    """
    A class for setting up working OneTick session. It keeps configuration files during
    the session and allows to manage them. When instance is out of scope, then instance
    cleans up config files and configuration.
    You can leave the scope manually with method :py:meth:`close`.
    Also, session is closed automatically if this object is used as a context manager.

    .. note::
        It is allowed to have only one alive session instance in the process.

    If you don't use Session's instance, then ``ONE_TICK_CONFIG`` environment variable
    should be set to be able to work with OneTick.

    If config file is not set then temporary is generated.
    Config includes locator and acl file, and if they are not set, then they are generated.

    Parameters
    ----------
    config : str, :py:class:`onetick.py.session.Config`, optional
        Path to an existing OneTick config file; if it is not set, then config will be generated.
        If config is not set, then temporary config is generated. Default is None.
    clean_up : bool, optional
        A flag to control cleaning up process: if it is True then all temporary generated files will
        be automatically removed. It is helpful for debugging. The flag affects only generated files, but
        does not externally passed.

        By default,
        :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
    copy : bool, optional
        A flag to control file copy process: if it is True then all externally passed files will be
        copied before usage, otherwise all modifications during an existing session happen directly with
        passed config files. NOTE: we suggest to set this flag only when you fully understand it's effect.
        Default is True.
    override_env : bool, optional
        If flag is True, then unconditionally ``ONE_TICK_CONFIG`` environment variable will be overridden
        with a config that belongs to a Session. Otherwise ``ONE_TICK_CONFIG``
        will be defined in the scope of session only when it is not defined externally.
        For example, it is helpful when you test ascii_loader that uses 'ONE_TICK_CONFIG' only.

        Default is False ( default is False, because overriding external environment variable
        might be not obvious and desirable )
    redirect_logs: bool, optional
        If flag is True, then OneTick logs  will be redirected into a temporary log file. Otherwise
        logs will be mixed with output. Default is True.
    gather_performance_metrics: bool, optional
        If flag is True, then enables performance metrics gathering, by setting ``DUMP_PERF_METRICS`` config parameter.
        Sets ``redirect_logs`` flag to ``True``.

        Metrics are available after closing a session via ``session.performance_metrics`` property.

        .. warning::
            Due to current limitations, metrics are cumulative. So if you run multiple queries in the same process
            (even with different session objects), you'll get metrics for the whole process since it's start
            till end of a session with ``gather_performance_metrics=True``.

            To avoid this, you can create a session in a separate process, either by using Python's ``multiprocessing``
            or by moving required code to a separate Python script and running it in a new process.

        .. note::
            Metrics are gathered for all operations in the session between its creation and closing.

    Examples
    --------

    If session is defined with environment, OneTick can be used right away:

    >>> 'ONE_TICK_CONFIG' in os.environ
    True
    >>> list(otp.databases()) # doctest: +ELLIPSIS
    ['COMMON', 'DEMO_L1', ..., 'SOME_DB', 'SOME_DB_2'...
    >>> data = otp.DataSource('SOME_DB', symbol='S1', tick_type='TT')
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    Collecting performance metrics with ``gather_performance_metrics`` parameter:

    >>> with otp.Session(gather_performance_metrics=True) as session:  # doctest: +SKIP
    >>>    data_a = otp.DataSource('DB_A', symbol='S1', tick_type='TT')  # doctest: +SKIP
    >>>    data_b = otp.DataSource('DB_B', symbol='S1', tick_type='TT')  # doctest: +SKIP
    >>>    _ = otp.run(otp.merge([data_a, data_b]))  # doctest: +SKIP
    >>>
    >>> session.performance_metrics  # doctest: +SKIP
    {
        'user_time': {'name': 'User Time', 'value': 3.39063, 'units': 's'},
        'system_time': {'name': 'System Time', 'value': 1.07813, 'units': 's'},
        'elapsed_time': {'name': 'Elapsed Time', 'value': 6.78816, 'units': 's'},
        'virtual_memory': {'name': 'Virtual Memory', 'value': 6261944320, 'units': 'bytes'},
        'virtual_memory_peak': {'name': 'Virtual Memory Peak', 'value': 6271926272, 'units': 'bytes'},
        'working_set': {'name': 'Working Set', 'value': 228110336, 'units': 'bytes'},
        'working_set_peak': {'name': 'Working Set Peak', 'value': 228126720, 'units': 'bytes'},
        'disk_read': {'name': 'Disk Read', 'value': 32289438, 'units': 'bytes'},
        'disk_write': {'name': 'Disk Write', 'value': 172906, 'units': 'bytes'}
    }
    """
    # TODO: create article for Session in Guides or Concepts

    _instance = None

    def __init__(
        self, config=None, clean_up=utils.default, copy=True, override_env=False, redirect_logs=True,
        gather_performance_metrics=False,
    ):
        self._construct(config, clean_up, copy, override_env, redirect_logs, gather_performance_metrics)

    def _construct(
        self, config=None, clean_up=utils.default, copy=True, override_env=False, redirect_logs=True,
        gather_performance_metrics=False,
    ):

        if Session._instance:
            raise MultipleSessionsException(
                "It is forbidden to use multiple active sessions simultaniously in one process"
            )

        def onetick_cfg_rollback(var):
            """
            function to rollback ONE_TICK_CONFIG state
            """

            def _impl():
                if var is None:
                    if "ONE_TICK_CONFIG" in os.environ:
                        del os.environ["ONE_TICK_CONFIG"]
                else:
                    os.environ["ONE_TICK_CONFIG"] = var

            return _impl

        self._lib = None
        self._env_rollback = onetick_cfg_rollback(os.environ.get("ONE_TICK_CONFIG", None))
        self._override_env = override_env

        self._config = Config.build(config, clean_up=clean_up, copy=copy, session_ref=self)
        # it is used in onetick-py-test
        os.environ["ONE_TICK_SESSION_CFG_PATH"] = self._config.path

        try:
            if "ONE_TICK_CONFIG" not in os.environ:
                os.environ["ONE_TICK_CONFIG"] = self._config.path
            else:
                if override_env:
                    os.environ["ONE_TICK_CONFIG"] = self._config.path
                else:
                    warnings.warn(
                        UserWarning(
                            "ONE_TICK_CONFIG env variable has been set before a session, "
                            "and in the session scope it is not related to the session config. "
                            "If you want to make ONE_TICK_CONFIG env variable be consistent "
                            "with the session, then look at the override_env flag "
                            "for the Session constructor"
                        )
                    )

            otli.OneTickLib().cleanup()

            self._performance_metrics_parser = None
            if gather_performance_metrics:
                self._performance_metrics_parser = PerformanceMetricsParser()
                utils.modify_config_param(
                    self._config.path, "DUMP_PERF_METRICS", "ON", throw_on_missing=False,
                )
                redirect_logs = True

            self._log_file = log_file = None
            if redirect_logs:
                self._log_file = utils.TmpFile(suffix=".onetick.log", clean_up=clean_up)
                log_file = self._log_file.path

            self._lib = otli.OneTickLib(self._config.path, log_file=log_file)
        except Exception:
            self._env_rollback()
            # TODO: rollback, but need to wait BDS-91
            raise

        # force reload cfg/locator/acl, because it is not reloaded after each time session if re-created
        if os.getenv('OTP_WEBAPI_TEST_MODE'):
            utils.reload_config(None, config_type='MAIN_CONFIG')
            self.locator.reload()
            self.acl.reload()

        self._ts_dbs = {}

        # PY-1352: this line must be the last
        # in case we get exception anywhere before, we shouldn't set Session class variable,
        # because it will affect the creation of all the future Session objects
        Session._instance = self

    def use(self, *items):
        """
        Makes DB or TS available inside the session.

        Parameters
        ----------
        items : :py:class:`~onetick.py.DB` or :py:class:`~onetick.py.servers.RemoteTS` objects
            Items to be added to session.

        Examples
        --------

        (note that ``session`` is created before this example)

        >>> list(otp.databases()) # doctest: +ELLIPSIS
        ['COMMON', 'DEMO_L1', ...]
        >>> new_db = otp.DB('ZZZZ')
        >>> session.use(new_db)
        >>> list(otp.databases()) # doctest: +ELLIPSIS
        ['COMMON', 'DEMO_L1', ..., 'ZZZZ']
        """
        self.locator.add(*items)
        dbs = []
        for item in items:
            if isinstance(item, _db.db._DB):
                dbs.append(item)
        try:
            if dbs:
                self.acl.add(*dbs)
        except Exception:
            self.locator.remove(*items)
            raise

    def use_stub(self, stub_name):
        """
        Adds stub-DB into the session.
        The shortcut for ``.use(otp.DB(stub_name))``

        Parameters
        ----------
        stub_name : str
            name of the stub
        """
        return self.use(_db.DB(stub_name))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            # session will be closed due to exception, no need to parse performance metrics
            self._performance_metrics_parser = None

        self.close()

    @staticmethod
    def _available_dbs():
        return _databases()

    def _get_ts_dbs(self):
        locator_dbs = self.locator.databases

        all_dbs = self._available_dbs()

        for db_name in all_dbs:
            if db_name not in locator_dbs and db_name not in self._ts_dbs:
                self._ts_dbs[db_name] = _db.db._DB(db_name)

    def close(self):
        """
        Close session
        """
        if Session._instance == self:
            try:
                if self._config:
                    del self._config
                    self._config = None

            finally:
                if self._lib:
                    self._lib.cleanup()
                    self._lib = None

                self._env_rollback()

                Session._instance = None

                if self._performance_metrics_parser:
                    self._performance_metrics_parser.parse(self._log_file)

    def __del__(self):
        self.close()

    @property
    def config(self):
        """
        A reference to the underlying Config object that represents OneTick config file.

        Returns
        -------
        :py:class:`onetick.py.session.Config`
        """
        return self._config

    @config.setter
    def config(self, cfg):
        self.close()
        self._construct(cfg, override_env=self._override_env)

    @property
    def acl(self):
        """
        A reference to the underlying ACL object that represents OneTick access control list file.

        Returns
        -------
        :py:class:`onetick.py.session.ACL`
        """
        return self._config.acl

    @property
    def locator(self):
        """
        A reference to the underlying Locator that represents OneTick locator file.

        Returns
        -------
        :py:class:`onetick.py.session.Locator`
        """
        return self._config.locator

    @property
    def license(self):
        return self._config.license

    @property
    def ts_databases(self):
        self._get_ts_dbs()
        return self._ts_dbs

    @property
    def databases(self):
        return self._available_dbs()

    @property
    def performance_metrics(self):
        if self._instance:
            raise RuntimeError("Trying to get performance metrics before closing the session.")

        if not self._performance_metrics_parser:
            raise RuntimeError(
                "Trying to get performance metrics, "
                "however session was created without `gather_performance_metrics=True`."
            )

        return self._performance_metrics_parser.metrics


class TestSession(Session):
    def __init__(self, *args, **kwargs):
        """
        This class does the same as :py:class:`onetick.py.session.Session`,
        but also defines default required :py:attr:`otp.config <onetick.py.configuration.Config>` values.

        Using this session object is the equivalent of defining these configuration values:

        ::

            otp.config['tz'] = 'EST5EDT'
            otp.config['default_db'] = 'DEMO_L1'
            otp.config['default_symbol'] = 'AAPL'
            otp.config['default_start_time'] = datetime(2003, 12, 1, 0, 0, 0)
            otp.config['default_end_time'] = datetime(2003, 12, 4, 0, 0, 0)

        ``DEMO_L1`` is the database defined in the default locator generated by onetick.py
        and it is also a name of the database commonly used in the OneTick ecosystem, academy courses, etc.
        """
        configuration.config['tz'] = 'EST5EDT'
        configuration.config['default_db'] = 'DEMO_L1'
        configuration.config['default_symbol'] = 'AAPL'
        configuration.config['default_start_time'] = datetime(2003, 12, 1, 0, 0, 0)
        configuration.config['default_end_time'] = datetime(2003, 12, 4, 0, 0, 0)
        super().__init__(*args, **kwargs)


class HTTPSession:
    param_list = ['http_address', 'http_username', 'http_password', 'access_token', 'http_proxy', 'https_proxy']

    def __init__(self,
                 http_address,
                 http_username=None,
                 http_password=None,
                 access_token=None,
                 http_proxy=None,
                 https_proxy=None):
        """
        This class must be used only for WebAPI connection,
        to set HTTP connection parameters.
        """
        import onetick.py as otp  # noqa
        self._restore_config = {}
        for param in self.param_list:
            if locals()[param]:
                self._restore_config[param] = otp.config.get(param)
                otp.config.__setattr__(param, locals()[param])

    def close(self):
        # restore config
        import onetick.py as otp  # noqa
        for param, value in self._restore_config.items():
            otp.config.__setattr__(param, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from onetick.py.core.column import _Column
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def _add_prefix_and_suffix(
    self: 'Source',
    prefix='',
    suffix='',
    columns=None,
    ignore_columns=None,
) -> Optional['Source']:
    if not prefix and not suffix:
        raise ValueError('Both suffix and prefix are empty')
    if ' ' in prefix:
        raise ValueError(f'There is space in prefix: {prefix}')
    if ' ' in suffix:
        raise ValueError(f'There is space in suffix: {prefix}')
    columns = columns or []
    ignore_columns = ignore_columns or []
    if columns and ignore_columns:
        raise ValueError('It is allowed to use only one of `columns` or `ignore_columns` parameters at a time')
    schema = self.schema
    for column in columns:
        if column not in schema:
            raise AttributeError(f'Column `{column}` does not exist in the schema')
    for column_name in columns or schema:
        if column_name in ignore_columns:
            continue
        new_column_name = f'{prefix}{column_name}{suffix}'
        if new_column_name in self.__dict__:
            if (not columns and not ignore_columns
                    or columns and new_column_name in columns
                    or ignore_columns and new_column_name not in ignore_columns):
                # if the column with the same name already exists, but will be renamed too,
                # then we don't need to raise exception, as it will have a different name after renaming
                continue
            if prefix:
                raise AttributeError(f'Column {new_column_name} already exists, please, use another prefix')
            else:
                raise AttributeError(f'Column {new_column_name} already exists, please, use another suffix')
    for column_name in columns or schema:
        if column_name in ignore_columns:
            continue
        new_column_name = f'{prefix}{column_name}{suffix}'
        self.__dict__[column_name].rename(new_column_name, update_parent_object=False)
        self.__dict__[new_column_name] = self.__dict__[column_name]
        del self.__dict__[column_name]
    if not columns and not ignore_columns:
        self.sink(otq.RenameFieldsEp(rename_fields=f'(.*)={prefix}\\1{suffix}', use_regex=True))
    elif columns:
        renames = [f'{column}={prefix}{column}{suffix}' for column in columns]
        self.sink(otq.RenameFieldsEp(rename_fields=','.join(renames)))
    else:
        renames = [f'{column}={prefix}{column}{suffix}' for column in schema if column not in ignore_columns]
        self.sink(otq.RenameFieldsEp(rename_fields=','.join(renames)))
    return self


def _is_excluded(s: str, not_to_rename: List[str]) -> bool:
    for excluded in not_to_rename:
        if re.match(excluded, s):
            return True
    return False


def _get_columns_names_renaming(schema, rename_dict: Dict[str, str], not_to_rename: List[str]) -> Dict[str, str]:
    """
    We can't be sure python Source has consistent columns cache, because sinking complex event processors
    can change columns unpredictable, so if user will specify regex as a param, we will pass regex
    as an onetick's param, but rename all matched columns from python Source cache.

    Parameters
    ----------
    rename_dict:
        Dict old_name -> new_name. Some of the dictionary's items use regex.

    Returns
    -------
    output_dict:
        Dict old_name -> new_name used for renaming columns in Source. None of this dictionary's items use regex.
    """
    output_dict = {}
    for old_name, new_name in rename_dict.items():
        matching_columns = [col for col in schema if re.match(old_name, col) and not _is_excluded(col, not_to_rename)]
        for col in matching_columns:
            output_dict[col] = re.sub(old_name, new_name, col)
    return output_dict


@inplace_operation
def add_prefix(self: 'Source', prefix, inplace=False, columns=None, ignore_columns=None) -> Optional['Source']:
    """
    Adds prefix to all column names (except **TIMESTAMP** (or **Time**) special column).

    Parameters
    ----------
    prefix : str
        String prefix to add to all columns.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.
    columns: List[str], optional
        If set, only selected columns will be updated with prefix. Can't be used with ``ignore_columns`` parameter.
    ignore_columns: List[str], optional
        If set, selected columns won't be updated with prefix. Can't be used with ``columns`` parameter.

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Add prefix *test_* to all columns (note that column **Time** is not renamed):

    >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL')
    >>> data = data.add_prefix('test_')
    >>> otp.run(data, start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2))
                         Time  test_PRICE  test_SIZE
    0 2022-03-01 00:00:00.000         1.3        100
    1 2022-03-01 00:00:00.001         1.4         10
    2 2022-03-01 00:00:00.002         1.4         50

    Parameter ``columns`` specifies columns to be updated with prefix:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data = data.add_prefix('test_', columns=['A', 'B', 'C'])
    >>> otp.run(data)
            Time  test_A  test_B  test_C  D  E
    0 2003-12-01       1       2       3  4  5

    Parameter ``ignore_columns`` specifies columns to ignore:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data = data.add_prefix('test_', ignore_columns=['A', 'B', 'C'])
    >>> otp.run(data)
            Time  A  B  C  test_D  test_E
    0 2003-12-01  1  2  3       4       5

    Parameters ``columns`` and ``ignore_columns`` can't be used at the same time:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data.add_prefix('test_', columns=['B', 'C'], ignore_columns=['A'])
    Traceback (most recent call last):
        ...
    ValueError: It is allowed to use only one of `columns` or `ignore_columns` parameters at a time

    Columns can't be renamed if their resulting name will be equal to existing column name:

    >>> data = otp.Tick(X=1, XX=2)
    >>> data.add_prefix('X', columns=['X'])
    Traceback (most recent call last):
        ...
    AttributeError: Column XX already exists, please, use another prefix

    """
    return self._add_prefix_and_suffix(
        prefix=prefix,
        columns=columns,
        ignore_columns=ignore_columns,
    )


@inplace_operation
def add_suffix(self: 'Source', suffix, inplace=False, columns=None, ignore_columns=None) -> Optional['Source']:
    """
    Adds suffix to all column names (except **TIMESTAMP** (or **Time**) special column).

    Parameters
    ----------
    suffix : str
        String suffix to add to all columns.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.
    columns: List[str], optional
        If set, only selected columns will be updated with suffix. Can't be used with ``ignore_columns`` parameter.
    ignore_columns: List[str], optional
        If set, selected columns won't be updated with suffix. Can't be used with ``columns`` parameter.

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Add suffix *_test* to all columns (note that column **Time** is not renamed):

    >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL')
    >>> data = data.add_suffix('_test')
    >>> otp.run(data, start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2))
                         Time  PRICE_test  SIZE_test
    0 2022-03-01 00:00:00.000         1.3        100
    1 2022-03-01 00:00:00.001         1.4         10
    2 2022-03-01 00:00:00.002         1.4         50

    Parameter ``columns`` specifies columns to be updated with suffix:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data = data.add_suffix('_test', columns=['A', 'B', 'C'])
    >>> otp.run(data)
            Time  A_test  B_test  C_test  D  E
    0 2003-12-01       1       2       3  4  5

    Parameter ``ignore_columns`` specifies columns to ignore:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data = data.add_suffix('_test', ignore_columns=['A', 'B', 'C'])
    >>> otp.run(data)
            Time  A  B  C  D_test  E_test
    0 2003-12-01  1  2  3       4       5

    Parameters ``columns`` and ``ignore_columns`` can't be used at the same time:

    >>> data = otp.Tick(A=1, B=2, C=3, D=4, E=5)
    >>> data.add_suffix('_test', columns=['B', 'C'], ignore_columns=['A'])
    Traceback (most recent call last):
        ...
    ValueError: It is allowed to use only one of `columns` or `ignore_columns` parameters at a time

    Columns can't be renamed if their resulting name will be equal to existing column name:

    >>> data = otp.Tick(X=1, XX=2)
    >>> data.add_suffix('X', columns=['X'])
    Traceback (most recent call last):
        ...
    AttributeError: Column XX already exists, please, use another suffix
    """
    return self._add_prefix_and_suffix(
        suffix=suffix,
        columns=columns,
        ignore_columns=ignore_columns,
    )


@inplace_operation
def rename(self: 'Source', columns=None, use_regex=False, fields_to_skip=None, inplace=False) -> Optional['Source']:
    r"""
    Rename columns

    Parameters
    ----------
    columns : dict
        Rules how to rename in the following format: {<column> : <new-column-name>},
        where <column> is either existing column name of str type or reference to a column,
        and <new-column-name> a new column name of str type.
    use_regex: bool
        If true, then old-name=new-name pairs in the `columns` parameter are treated as regular expressions.
        This allows bulk renaming for field names. Notice that regular expressions for old names are treated as
        if both their prefix and their suffix are .*, i.e. the prefix and suffix match any substring.
        As a result, old-name *XX* will match all of *aXX*, *aXXB*, and *XXb*, when `use_regex=true`.
        You can have old-name begin from ^ to indicate that .* prefix does not apply,
        and you can have old name end at $ to indicate that .* suffix does not apply.
        Default: false
    fields_to_skip: list of str
        A list of regular expressions for specifying fields that should be skipped
        (i.e., not be renamed). If a field is matched by one of the specified regular expressions,
        it won't be considered for renaming.
        Default: None
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise, method returns a new modified
        object.

    Returns
    -------
    :class:`Source` or ``None``

    See also
    --------
    **RENAME** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1], Y=[2])
    >>> data = data.rename({'X': 'XX',
    ...                     data['Y']: 'YY'})
    >>> otp.run(data)
            Time  XX  YY
    0 2003-12-01   1   2

    >>> data = otp.Tick(**{'X.X': 1, 'X.Y': 2})
    >>> data = data.rename({r'X\.(.*)': r'\1'}, use_regex=True)
    >>> otp.run(data)
            Time   X   Y
    0 2003-12-01   1   2

    >>> data = otp.Tick(**{'X.X': 1, 'X.Y': 2})
    >>> data = data.rename({r'X\.(.*)': r'\1'}, use_regex=True, fields_to_skip=['X.Y'])
    >>> otp.run(data)
            Time   X X.Y
    0 2003-12-01   1   2
    """
    if columns is None:
        columns = {}

    # prepare
    items = {}
    out_names = set()
    fields_to_skip = fields_to_skip or []

    for in_obj, out_obj in columns.items():
        if isinstance(in_obj, _Column):
            items[in_obj.name.strip()] = out_obj.strip()
        elif isinstance(in_obj, str):
            items[in_obj.strip()] = out_obj.strip()
        else:
            raise TypeError(f"It is not supported to rename item '{in_obj}' of type {type(in_obj)}'")

        if out_obj in out_names:
            raise AttributeError(
                f"You want to rename '{in_obj}' into '{out_obj}', "
                f"but also want to rename another column into '{out_obj}'"
            )

        out_names.add(out_obj)

    schema_update_dict = items
    if use_regex:
        schema_update_dict = _get_columns_names_renaming(self.schema, items, fields_to_skip)

    # validate
    for in_key, out_key in schema_update_dict.items():
        if in_key not in self.__dict__ or not isinstance(self.__dict__[in_key], _Column):
            raise AttributeError(f"There is no '{in_key}' column to rename")

        if out_key in self.__dict__ and isinstance(self.__dict__[out_key], _Column):
            raise AttributeError(f"Column '{out_key}' already exists")

    # apply
    for in_key, out_key in schema_update_dict.items():
        self.__dict__[in_key].rename(out_key, update_parent_object=False)
        self.__dict__[out_key] = self.__dict__[in_key]
        del self.__dict__[in_key]

    rename_rules = [key + "=" + value for key, value in items.items()]
    kwargs: Dict[str, Any] = dict(rename_fields=",".join(rename_rules))
    if use_regex:
        kwargs['use_regex'] = True
    if fields_to_skip:
        kwargs['fields_to_skip'] = ",".join(fields_to_skip)

    self.sink(otq.RenameFieldsEp(**kwargs))

    return self



from onetick.py.otq import otq
from onetick.py import state
from onetick.py.core._internal._op_utils.is_const import is_const

from onetick.py.core.column_operations.base import _Operation
from onetick.py.core._internal._state_objects import _StateColumn, _StateBase, _TickSequence
from onetick.py.types import type2str, value2str


class StateVars:
    def __init__(self, owner, previous=None):
        """ Class for organizing state variables

        Parameters
        ----------
        owner: _Source
            Owner of the object
        previous: collections of dict of _Source, optional
            Dictionaries to copy state variables from
        """
        self._owner = owner
        self._columns = {}
        previous = previous if isinstance(previous, (list, tuple)) else (previous,)
        for p in previous:
            if p:
                for name, column in p.state_vars.items:
                    value = column
                    if isinstance(value, _StateBase):
                        value = value.copy(obj_ref=self, name=name)
                    self._columns[name] = value

    def __contains__(self, item):
        return item in self._columns

    def __getitem__(self, item):
        return self._columns[item]

    def __setitem__(self, key, value):
        if isinstance(value, _Operation):
            value = self._set_var_from_operation(key, value)
        elif isinstance(value, _StateBase):
            value = value.copy(obj_ref=self, name=key)
        elif key not in self._columns:
            value = state.var(value).copy(obj_ref=self, name=key)

        if key in self._columns:
            self._owner._update_field(self._columns[key], value)
        else:
            self._declare_new_var(key, value)
            self._columns[key] = value

    def _set_var_from_operation(self, key, value):
        if isinstance(value, _StateColumn):
            assert not value.obj_ref, "You can't set state variable of one Source to another"
            value.obj_ref = self._owner
            value.name = key
        else:
            if not is_const(value):
                assert (key in self._columns), "OneTick supports state variables creation with constants only"
            else:
                value = state.var(value).copy(obj_ref=self, name=key)
        return value

    @property
    def items(self):
        return tuple(self._columns.items())  # prevent mutation

    @property
    def names(self):
        return tuple(self._columns.keys())

    def _declare_new_var(self, name, value):
        # obj_ref is None means that we need to declare it
        base_type = value.dtype
        str_type = ""

        if isinstance(value, _TickSequence):
            # TODO: maybe create definition() for _StateColumn too and move this logic there
            str_type = value._definition()
        else:
            str_type = type2str(base_type)

        expression = f'{str_type} {name}'

        import onetick.py as otp
        if value.default_value is not None:
            if isinstance(value.default_value, otp.Source):
                value_expression = otp.eval(value.default_value).to_eval_string(self._owner._tmp_otq)
            else:
                # TODO: PY-952: use to_eval_string(self._owner._tmp_otq) here if otp.eval is passed
                value_expression = value2str(value.default_value)
            expression += f' = {value_expression}'
        self._owner.sink(otq.DeclareStateVariables(variables=expression, scope=value.scope))
